# coding=utf-8

''' tagsPlorer library  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

import logging, os, pickle, sys, zlib
from functools import reduce

from tagsplorer.constants import ALL, COMB, CONFIG, DOT, FROM, GLOBAL, IGNFILE, IGNORE, IGNORED, IGNOREDS, ON_WINDOWS, PICKLE_PROTOCOL, SEPA, SKIP, SKIPD, SKIPDS, SKPFILE, SLASH, ST_SIZE, TAG, TOKENIZER
from tagsplorer.utils import anyParentIsSkipped, appendnew, dd, dictGet, dictGetSet, findIndexOrAppend, getTsMs, isDir, isFile, isGlob, lappend, normalizer, pathHasGlobalIgnore, pathHasGlobalSkip, pathNorm, safeSplit, sjoin, splitByPredicate, wrapExc, xall, xany


_log = logging.getLogger(__name__)
def log(func): return (lambda *s: func(sjoin([_() if callable(_) else _ for _ in s]), **({"stacklevel": 2} if sys.version_info >= (3, 8) else {})))
debug, info, warn, error = log(_log.debug), log(_log.info), log(_log.warning), log(_log.error)


class ConfigParser(object):  # TODO is there a faster serialization protocol?
  ''' Much simplified ini-style config file. No line continuation, no colon separation, no symbolic replacement, no comments, no empty lines. '''

  def __init__(_): _.sections = {}  # there is always a global section [] and optionally exactly one section per root-relative path

  def load(_, fd):
    ''' Load the .cfg file.
        fd: file-like object to enable reading from arbitrary data source and position in resource (e.g. after consuming the timestamp)
        returns: only the global section []
    '''
    title = ''; section = dd()
    for line in fd.readlines():
      line = line.strip()  # just in case of incorrect formatting IDEA could be optimized away, or comments allowed
      if line.startswith('['):  # new section detected: before processing, store the last section
        if len(section): _.sections[title] = section  # stores empty global section when encountering beginning of first (and even global) section
        title = line[1:line.index(']')]; section = dd()  # for the next section
      elif line != '':  # content found
        try:
          idx = line.index('=')  # try to parse
          key, value = line[:idx].lower(), line[idx+1:]  # HINT: was .lower().rstrip(), but when accessing only via this script's API there's no need for that
          if value != '' and (key in [TAG, FROM]\
          or title == "" and  key in [SKIPD, IGNORED, GLOBAL]): section[key].append(value)  # allows several values per key
          elif key in [IGNORE, SKIP] and key not in section:    section[key] = None  # only keep key as marker
          else: warn(f"Encountered illegal configuration key '{key}'. Skip entry")
        except: warn(f"Key without value for illegal key '{line}'")
      else: break  # an empty line terminates file
    if len(section): _.sections[title] = section  # store last section
    return {k: v if v.strip().lower() not in ("true", "false") else v.strip().lower() == "true" for k, v in (wrapExc(lambda: kv.split("=")[:2], lambda: (kv, None)) for kv in _.sections.get("", {}).get(GLOBAL, []))}

  def store(_, fd, parent = None):
    ''' Store the .cfg file.
        fd: file-like object to allow for writing additional contents (e.g. the timestamp)
        parent: optional global settings
    '''
    if parent is not None:  # store parent's properties
      if "" not in _.sections: _.sections[""] = dd()
      _.sections[""][GLOBAL] = sorted([f"{k.lower()}={parent.__dict__[k]}" for k, v in (kv.split("=")[:2] for kv in _.sections.get("", {}).get(GLOBAL, []))])  # store updated config
    for title, _map in sorted(_.sections.items()):  # ordered is better for VCS
      if len(_map) == 0 or xall(lambda v: len(v) == 0, _map.values()): continue
      fd.write(f"[{title}]\n")
      for key, values in sorted(_map.items()):  # no need for iteritems, as supposedly small (max. size = 4)
        if values is None: fd.write(f"{key}=\n")  # skip or ignore
        else:
          for value in sorted(values): fd.write(f"{key}={value}\n")  # tag or from
    fd.write("\n")  # termination of config contents


class Configuration(object):
  ''' Contains all tag configuration. '''

  def __init__(_, case_sensitive = None):
    _.paths = {}  # {relative dir path -> {marker -> [entries]}}
    _.reset(case_sensitive)
    normalizer.setupCasematching(_.case_sensitive, suppress = True)  # initialize normalizer, but can be overridden during load()

  def reset(_, case_sensitive = None):
    _.case_sensitive = (not ON_WINDOWS) if case_sensitive is None else case_sensitive  # search behavior
    _.reduce_storage = False           # storage behavior
    _.compression = 2                  # good fast compromise: uncompressed pickling is faster than any bz2 compression, but zlib level 2 seems to get best trade-off. 0 means uncompressed

  def logConfiguration(_):
    ''' Display debug info. '''
    info("Configuration:  " + "  ".join(f"{k}: %s" % ("On" if v else "Off") for k, v in [
        ("case_sensitive",  _.case_sensitive),
        ("reduce_storage",  _.reduce_storage),
        ("on_windows",      ON_WINDOWS)
      ]))

  def load(_, folder, index_ts = None):
    ''' Load configuration from file, unless up to date.
        folder:   the configuration folder to load from
        index_ts: timestamp from inside the index file, or None (force load)
        returns:  False if index is still current, otherwise True (configuration loaded and a new index must be created)
    '''
    with open(os.path.join(folder, CONFIG), 'r', encoding = "utf-8") as fd:
      timestamp = float(fd.readline().rstrip())
      if index_ts and timestamp == index_ts:
        debug("Index is up to date")  # skip loading configuration
        _.logConfiguration()
        return False  # no skew detected, allow using old index's interned configuration ("self" will be discarded)
      debug(f"Load configuration from {folder}{'' if not index_ts else ' because index is outdated'}")
      cp = ConfigParser(); dat = cp.load(fd); _.__dict__.update(dat)  # update Configuration() object with loaded global options
      _.paths = cp.sections
      normalizer.setupCasematching(_.case_sensitive)  # update with just loaded setting
      _.logConfiguration()
      return True

  def store(_, folder, timestamp = None):
    ''' Store configuration to file, prepending data by the timestamp, or the current time. '''
    debug(f"Store configuration to {folder}%s" % (" (%.1f)" % timestamp if timestamp else ""))
    if not timestamp: timestamp = getTsMs()  # for those cases, in which we modify only the config file (e.g. tag, untag, config)
    cp = ConfigParser(); cp.sections = _.paths
    with open(os.path.join(folder, CONFIG), "w", encoding = "utf-8") as fd: fd.write(f"{timestamp}\n"); cp.store(fd, parent = _)
    info(f"Wrote {os.stat(os.path.join(folder, CONFIG))[ST_SIZE]} config bytes")

  def addTag(_, folder, tag, poss, negs, force = False):
    ''' For a given folder, add a tag for inclusive and exclusive conjunctive globs.
        folder: root-relative folder to operate on (section key in the configuration)
        tag: the tag to add to the configuration for the given folder
        poss: list of inclusive globs to apply the tag to
        negs: list of exclusive globs to apply the tag not to
        force: allows adding tags already covered inclusively or exclusively by existing globs TODO reimplement safety checks
        returns: was successfully added?
    '''
    debug(f"addTag +%s -%s to {folder}/{tag}%s" % (COMB.join(poss), COMB.join(negs), " force" if force else ""))
    conf = dictGet(_.paths, folder, {})
    to_add = f"{tag};%s;%s" % (SEPA.join(sorted(poss)), SEPA.join(sorted(negs)))
    for line in conf.get(TAG, []):  # all pattern markers for the given folder
      if line.strip() == to_add:
        warn(f"Tag <{tag}> in {folder} for +%s -%s already defined, skip" % (COMB.join(sorted(poss)), COMB.join(sorted(negs))))
        return False

    info(f"Tag <{tag}> in '{folder}' for +%s -%s" % (COMB.join(poss), COMB.join(negs)))
    conf = dictGetSet(_.paths, folder, {})  # creates empty config entry if it doesn't exist
    dictGetSet(conf, TAG, []).append(to_add)
    return True

  def delTag(_, folder, tag, poss, negs):
    ''' For a given folder, remove a tag for inclusive and exclusive conjuntive globs.
        folder: root-relative folder to operate on (section key in the configuration)
        tag: the tag to remove from the configuration for the given folder
        poss: list of inclusive globs to remove the tag from
        negs: list of exclusive globs to remove the tag from
        returns: was successfully removed?
    '''
    debug(f"delTag +%s -%s from {folder}/{tag}" % (COMB.join(poss), COMB.join(negs)))
    conf = dictGet(_.paths, folder, {})  # no need to create if missing, since we are removing anyway
    to_del = f"{tag};%s;%s" % (SEPA.join(sorted(poss)), SEPA.join(sorted(negs)))
    found = False
    for line in conf.get(TAG, []):
      if line.strip() == to_del: found = True
    if found:
      warn(f"Untag '{tag}' in {folder} for +%s -%s" % (COMB.join(sorted(poss)), COMB.join(sorted(negs))))
      conf.get(TAG)[:] = [entry for entry in conf.get(TAG) if entry != to_del]  # remove matches
    return found

  def showTags(_, folder):
    ''' For a given folder, show all defined tags.
        folder: root-relative folder to operate one
    '''
    debug(f"showTags for {folder}")
    inPath = set(safeSplit(folder, SLASH))  # break path into constituents
    inPath.update(reduce(lambda prev, step: prev + TOKENIZER.split(step) + TOKENIZER.split(normalizer.filenorm(step)), inPath, []))
    warn("Tags derived from path: " + COMB.join(inPath))  # TODO filter output by ignore and skip
    conf = dictGet(_.paths, folder, {})
    tags = dd()
    for line in conf.get(TAG, []): tags[line.split(SEPA)[0]].append(line)
    for key, defs in sorted(tags.items()): print("\n".join(defs))


class Indexer(object):
  ''' Main index creation. Walks through file tree and indexes folder tags.
      Addtionally, tags for single files or globs, and those mapped by FROM markers are included in the index.
      File-specific tags are only determined during the find phase.
      TODO test and benchmark other methods of serialization, e.g. shelve, pickleDB and others
  '''

  def __init__(_, startDir):
    ''' startDir: absolute path. '''
    _.timestamp = 1.23456789  # for comparison with configuration timestamp
    _.cfg = None  # reference to latest corresponding configuration
    _.root = pathNorm(startDir[:-1] if startDir.endswith(":\\") else startDir)  # slash-normalized absolute path
    _.tagdirs = []  # array of tags (plain dirnames and manually set tags (not represented in parent), both case-normalized and as-is)
    _.tagdir2parent = []  # index of dir entry (tagdirs) -> index of parent to represent tree structure (excluding folder links)
    _.tagdir2paths = dd()  # dirname/tag index -> list of [all path indices relevant for that dirname/tag]

  def load(_, filename, ignore_skew = False, recreate_index = False):
    ''' Load a pickled index into memory. Optimized for speed.
        filename: absolute path to the index file
        ignore_skew:    if True, ignore the fact that index and config timestamps deviate. used in tests and for --keep-index
        recreate_index: if True, create a new index even if timestamps still match
    '''
    debug(f"load('{filename}': ignore_skew %s, recreate_index %s)" % ("Yes" if ignore_skew else "No", "Yes" if recreate_index else "No"))
    try: del _.allPaths  # cache flag used when run as a web server
    except AttributeError: pass  # delete cache when reloading
    with open(filename, "rb") as fd:
      info("Read index from " + filename)
      c = pickle.loads(wrapExc(lambda: zlib.decompress(fd.read()), fd.read))
      _.cfg, _.timestamp, _.tagdirs, _.tagdir2parent, _.tagdir2paths = c.cfg, c.timestamp, c.tagdirs, c.tagdir2parent, c.tagdir2paths
      cfg = Configuration(_.cfg.case_sensitive)
      if (recreate_index or cfg.load(os.path.dirname(os.path.abspath(filename)), _.timestamp)) and not ignore_skew:
        info("Recreate index considering configuration")
        _.cfg = cfg  # set more recent config and update
        _.walk()  # re-create this index
        _.store(filename)
      normalizer.setupCasematching(_.cfg.case_sensitive)  # update with just loaded setting

  def store(_, filename, config_too = True):
    ''' Persist index in a file, including currently active configuration. '''
    with open(filename, "wb") as fd:
      nts = getTsMs()
      _.timestamp = _.timestamp + 0.001 if nts <= _.timestamp else nts  # assign new date, ensure always differing from old value
      debug("Store index to " + filename)
      fd.write(zlib.compress(pickle.dumps(_, protocol = PICKLE_PROTOCOL), _.cfg.compression) if _.cfg.compression else pickle.dumps(_, protocol = PICKLE_PROTOCOL))
      if config_too:
        debug("Update configuration to match new index timestamp")
        _.cfg.store(os.path.dirname(os.path.abspath(filename)), _.timestamp)  # update timestamp in configuration
    info(f"Wrote {os.stat(filename)[6]} index bytes ({len(_.tagdirs)} entries and %d paths)" % (sum([len(p) for p in _.tagdir2paths])))

  def walk(_, cfg = None):
    ''' Build index by recursively traversing the folder tree.
        cfg: if set, use that configuration instead of the one in the root.
    '''
    info("Walk folder tree to update index")
    if cfg: _.cfg = cfg
    if _.cfg is None: raise Exception("No configuration loaded. Cannot traverse folder tree")
    debug(f"Configuration: case_sensitive = {_.cfg.case_sensitive}, reduce_storage = {_.cfg.reduce_storage}")
    _.tagdirs = [""]       # list of directory names, duplicates allowed and required to represent the tree structure. "" represents the root folder
    _.tagdir2parent = [0]  # pointer to parent directory. the self-reference at index 0 marks root. each index position responds to one entry in tagdirs (!)
    _.tagdir2paths = dd()  # maps index of tagdir entries to list of path leaf indexes, to find all paths ending in that suffix
    _.tags = []            # temporary data structure for "set" of (manually set or folder-derived) tag names and file extensions, which gets mapped into the tagdirs structure
    _.tag2paths = dd()     # temporary data structure for config-specified tag and extension mapping to matching respective paths
    _._walk(_.root, 0)     # recursive indexing
    _.mapTagsIntoDirsAndCompressIndex()  # manual tags are combined in the index with folder names for simple lookup and filtering
    del _.tags, _.tag2paths  # remove temporary structures
    _.tagdir2paths = [_.tagdir2paths[i] if i in _.tagdir2paths else [] for i in range(max(_.tagdir2paths) + 1)] if _.tagdir2paths else []  # safely convert map values to list positions (as we don't know if all exist)
    info(f"Indexed {len(_.tagdirs)} folders with {len(_.tagdir2paths)} tags")

  def _walk(_, folder, findex, tags = None, last = 0):
    ''' Recursive traversal through folder tree.
        This will index the folder 'folder' and then recurse into its children folders.
        The function adds all parent folder names (up to root) plus the current directory's name as tags, unless told to ignore the current name by a configuration setting or file marker.
        The general approach is to add children's tags in the parent, then recurse, potentially removing added tags in the recursion if necessary.
        folder:  absolute folder path to add to the index (slash-normalized to one preceding and no trailing forward slash)
        findex:  specified folder's index number in the data structures (parent for the children processed here = current folder)
        tags:    list of aggregated tag indexes (folder names) of parent directories
        last:    number of last indexes used to store the current folder (1 if normalized only ot reduced storage, otherwise 2, 0 if ignored)
    '''
    debug(f"_walk '{folder}' findex {findex} {last} {tags}")
    if tags is None: tags = []  # because using default `= []` is a bad idea in Python
    ignore = False  # marks folder as "no tagging for this specific folder", according to local or global settings
    adds = set()    # for display only: indexes of additional tags valid for the current folder only, not to be promoted to children calls

    # 1.  get folder configuration, if any
    marks = _.cfg.paths.get(folder[len(_.root):], {})  # contains configuration for current folder, if any
    # 1a. check skip or ignore flags from configuration
    if SKIP   in marks or xany(lambda ignore: normalizer.globmatch(folder[folder.rindex(SLASH) + 1:] if folder and SLASH in folder else folder, ignore), dictGet(dictGet(_.cfg.paths, '', {}), SKIPD, []) + SKIPDS):
      info(f"Skip '{folder[len(_.root):]}' due to " + ('path skip' if SKIP in marks else 'global folder name skip'))
      return  # completely ignore sub-tree and break recursion
    if IGNORE in marks or xany(lambda ignore: normalizer.globmatch(folder[folder.rindex(SLASH) + 1:] if folder and SLASH in folder else folder, ignore), dictGet(dictGet(_.cfg.paths, '', {}), IGNORED, []) + IGNOREDS):
      info(f"Ignore '{folder[len(_.root):]}' due to " + ('path ignore' if IGNORE in marks else 'global folder name ignore'))
      ignore = True  # ignore this directory as a tag, and don't index its contents, but still continue recursion
    # 1b. read configured additional tags for folder and folder mapping from configuration into "tags" and "adds"
    elif TAG in marks:  # neither SKIP nor IGNORE in config: consider manual folder tagging
      for t in marks[TAG]:
        tag, pos, neg = t.split(SEPA)  # tag name, includes, excludes
        info(f"Tag <{tag}> (+<{pos}> -<{neg}>) in '{folder[len(_.root):]}'")
        i = findIndexOrAppend(_.tags, tag); adds.add(i)  # find existing index of that tag, or create return new index
        appendnew(_.tag2paths[i], findex)  # add tag and create link to the current folder ("tag" is a match for the current folder?)
    if FROM in marks:  # consider tags from mapped folders, even if proper folder is ignored
      for f in marks[FROM]:  # map configured tags, even if it contains an ignored marker
        info(f"Map from '{f}' into '{folder[len(_.root):]}'")
        other = os.path.normpath(os.path.join(folder, pathNorm(f)))[len(_.root):] if not f.startswith(SLASH) else pathNorm(f)
        _marks = _.cfg.paths.get(other, {})  # marks of mapped folder
        for t in _marks.get(TAG, []):
          tag, pos, neg = t.split(SEPA)  # HINT the actual pattern filtering is implemented in findFiles TODO or even findFolders
          info(f"Tag <{tag}> (+<{pos}> -<{neg}>) in '{folder[len(_.root):]}'")
          i = findIndexOrAppend(_.tags, tag); adds.add(i)
          appendnew(_.tag2paths[i], findex)

    # 2. process folder's file names (file extensions only)
    files, folders = splitByPredicate(os.listdir(folder), lambda f: isFile(folder + SLASH + f))  # HINT right-hand side is not automatically a directory, thus filtered below:
    folders[:] = sorted([f for f in folders if isDir(folder + SLASH + f)])  # remove special files like ".desktop"
    if SKPFILE in files:  # HINT allow other than lower case skip file? should be no problem, as even Windows allows lower-case file names and should match here
      info(f"Skip '{folder[len(_.root):]}' due to local skip marker file")
      return  # ignore entire sub-tree and break recursion
    if IGNFILE in files:
      info(f"Ignore '{folder[len(_.root):]}' due to local ignore marker file")
      ignore = True
    if not ignore:
      debug(f"Index file types for '{folder[len(_.root):]}'")
      for _file in (ff for ff in files if DOT in ff[1:]):  # index file extensions (including this in sub-folders), without propagation to sub-folders HINT dot-first files are ignored. see below
        ext = _file[_file.rindex(DOT):]  # split off extension, even when "empty" extension (filename ending in dot)
        if ext not in adds:
          i = findIndexOrAppend(_.tags, ext); adds.add(i)  # get or add file extension to local dir's tags only
          _.tag2paths[i].append(findex)  # add current dir to index of that extension
        iext = ext.lower()
        if iext != ext and not _.cfg.reduce_storage:  # store normalized extension if different from literal, unless told not to
          i = findIndexOrAppend(_.tags, iext); adds.add(i)  # add file extension to local dir's tags only
          _.tag2paths[i].append(findex)  # add current dir to index of that extension
      # TODO optionally also index file names and their tokens! this would also cover dot-first files ignored above

    # 3.  prepare recursion
    newtags = [t for t in tags[:-last if ignore else None]]  # tags to propagate into subfolders (except current folder names if ignored)
    cache = {}  # for faster path computation during output logging HINT cache was formerly in side folders loop
    for subfolder in ([] if ignore else folders):  # iterate sub-folders
      # 3a. add sub-folder name to "tagdirs" and "tagdir2parent"
      idxs  = []  # *first* element in "idx" is parent index to use in recursion for the currently processed subfolder
      addt  = set()  # per-subfolder local additional tokens
      added = 0

      info(f"Store literal folder '{subfolder}' for '{_.getPath(findex, cache)}'")
      idxs.append(len(_.tagdirs))  # for this subfolder, always add a *new* element, no matter if name already exists in the index, because parent differs (to keep tree structure)
      _.tagdirs.append(subfolder)  # now add the subfolder name to the list of known tags/folders (duplicates allowed, because we build the tree structure here)
      _.tagdir2parent.append(findex)  # placed at same index as tagdirs "next" position
      added += 1
      assert len(_.tagdirs) == len(_.tagdir2parent)  # invariant

      iname = subfolder.lower()
      if not _.cfg.reduce_storage and iname != subfolder:
        info(f"Store case-normalized folder '{iname}' for '{_.getPath(findex, cache)}'")
        idxs.append(len(_.tagdirs))     # new index
        _.tagdirs.append(iname)         # add to both
        _.tagdir2parent.append(findex)  # data structures
        added += 1
        assert len(_.tagdirs) == len(_.tagdir2parent)  # invariant

      tokens = [r for r in TOKENIZER.split(subfolder) if r not in ("", subfolder)]  # in addition to the folder parent mapping, split folder name into tokens
      for token in set(tokens):
        debug(f"Store literal token '{token}' for '{_.getPath(findex, cache)}'")
        i = findIndexOrAppend(_.tags, token); addt.add(i)
        _.tag2paths[i].append(idxs[-added])  # link token index to stored path constituent

        itoken = token.lower()
        if not _.cfg.reduce_storage and itoken != token:
          debug(f"Store case-normalized token '{itoken}' for '{_.getPath(findex, cache)}'")
          i = findIndexOrAppend(_.tags, itoken); addt.add(i)
          _.tag2paths[i].append(idxs[-added])
      assert len(_.tagdirs) == len(_.tagdir2parent)  # invariant

      if not ignore:  # then index current folder
        debug(f"Mark folder '{folder[len(_.root):]}{SLASH}{subfolder}' with <%s>" % (COMB.join(set([_.tagdirs[x] for x in (newtags + idxs)]) | set([_.tags[x] for x in (adds | addt)]))))  # per subfolder, not promoted to recursive call
        for tag in newtags + idxs:
          _.tagdir2paths[_.tagdirs.index(_.tagdirs[tag])].extend(idxs)  # add sub-folder reference(s) for all collected parent folder tags to the tag name

      # 4. recurse into subfolder
      _._walk(folder = folder + SLASH + subfolder, findex = idxs[-added], tags = newtags + ([] if ignore else idxs), last = added)  # recursion

  def mapTagsIntoDirsAndCompressIndex(_):
    ''' After recursion finished, map extensions or manually set tags into the tagdir structure to save space. '''
    info("Map tags into folder index")
    for itag, tag in enumerate(_.tags):
      idx = findIndexOrAppend(_.tagdirs, tag)  # get (first) index in list, or append if not found yet
      _.tagdir2paths[idx].extend(_.tag2paths[itag])  # tag2paths contains true parent folder index from _.tagdirs
    rm = [tag for tag, dirs in _.tagdir2paths.items() if len(dirs) == 0]  # find entries that are empty due to ignores
    for r in rm:
      del _.tagdir2paths[r]
      debug(f"Remove childless tag '{_.tagdirs[r]}'")  # remove empty lists from index (was too optimistic, removed by ignore or skip)
    debug(f"Removed {len(rm)} childless tags from index")
    found = 0
    for tag, dirs in _.tagdir2paths.items():
      _l = len(_.tagdir2paths[tag])  # get size of mapped paths
      _.tagdir2paths[tag] = set(dirs)  # remove duplicates
      found += (_l - len(_.tagdir2paths[tag]))  # check any removed
    if found: debug(f"Removed {found} duplicates from index")

  def getPath(_, idx, cache):
    ''' Return one root-relative path for the given index by recursively going through {index: name} mappings and combining them into a full path.
        idx:     folder entry index from _.tagdirs
        cache:   dictionary for speeding up consecutive calls (remembers already collected parent path mappings)
        returns: root-relative path string
    >>> i = Indexer("bla"); c = {}
    >>> i.tagdirs = ["", "a", "b", "c"]
    >>> i.tagdir2parent = [0,  0,   1,   1]  # same positions as in tagdirs
    >>> print(repr(i.getPath(0, c)))
    ''
    >>> print(c)
    {}
    >>> print(i.getPath(1, c))
    /a
    >>> print(c)
    {1: '/a'}
    >>> print(i.getPath(3, c))
    /a/c
    >>> print(sorted(c.items()))
    [(1, '/a'), (3, '/a/c')]
    '''
    if idx == 0: return ""  # root
    assert 0 <= idx < len(_.tagdirs)
    found = cache.get(idx, None)
    if found is not None: return found
    parent_idx = _.tagdir2parent[idx]  # get parent index from tree-structure
    parent = dictGetSet(cache, parent_idx, _.getPath(parent_idx, cache)) if parent_idx != 0 else ""  # recursive folder name resolution
    cache[idx] = new_ = parent + SLASH + _.tagdirs[idx]
    return new_

  def getPaths(_, ids, cache):
    ''' Returns a generator for respective paths of the given path index list.
        ids:     iterable of tagdirs ids
        cache:   dictionary for speeding up consecutive calls to _.getPath()
        returns: generator that yields root-relative path strings
    >>> i = Indexer("blupp")
    >>> i.tagdirs =       ["", "a", "b", "c"]
    >>> i.tagdir2parent = [0,  0,   1,   1]  # same positions as in tagdirs
    >>> c = {}; g = i.getPaths(range(4), c)
    >>> print(" ".join([next(g), next(g), next(g), next(g)]))  # indented because of empty root path
     /a /a/b /a/c
    >>> print(len(c))  # cache size
    3
    '''
    return (_.getPath(i, cache) for i in ids)

  def removeIncluded(_, includedTags, excludedPaths):
    ''' Return those paths, that have no manual tags or FROM tags from the inclusion list; subtract from the exclusion list to reduce set of paths to ignore.
        includedTags:  search tags to keep included (no extensions, no globs). Will be removed from the excludedPaths list
        excludedPaths: all paths for exclusive tags, scheduled for removal from search results
        returns:       list of paths to retain
        HINT: there is no need to check all tags in configuration, because they could be inclusive or exclusive, and are on a per-file basis
        the existence of a tag suffices for retaining paths
    '''
    retain = []
    for path in excludedPaths:
      conf = _.cfg.paths.get(path, {})
      if len(conf) == 0: retain.append(path); continue  # accept for removal, because no further tagging information exists
      tags, froms = conf.get(TAG, []), conf.get(FROM, [])
      for other in froms: tags.extend(_.cfg.paths.get(other, {}).get(TAG, []))
      retainit = True
      for value in tags:  # all explicit tags count, no matter if globs match anything or not (over-generic folder list)
        tg, inc, exc = value.split(SEPA)
        if tg in includedTags: retainit = False; break  # deny removal
      if retainit: retain.append(path)
    return set(retain)

  def findFolders(_, include, exclude, returnAll = False):
    ''' Find intersection of all indexed folders with specified tags, from over-generic index.
        include:   list of tag names that must be present
        exclude:   list of tag names that must not be present
        returnAll: shortcut flag that simply returns *all paths* from the index instead of finding and filtering results (from tp.find())
        returns:   list of folder paths (case-normalized or both normalized and as is, depending on the case-sensitive option)
    '''
    idirs, sdirs = dictGet(dictGet(_.cfg.paths, '', {}), IGNORED, []) + IGNOREDS, dictGet(dictGet(_.cfg.paths, '', {}), SKIPD, []) + SKIPDS  # get lists of ignored and skipped paths
    cache = {}
    def rebuild():
      debug(f"Build list of all paths.  Global ignores: {idirs}  Global skips: {sdirs}")
      return set(_.getPaths(list(reduce(lambda a, b: a.update(set(b)) or a, _.tagdir2paths, set())), cache))  # compute union of all paths (cached when running as a server)
    _.allPaths = wrapExc(lambda: _.allPaths, rebuild)  # lazy computation. fails and computes if property not defined
    alls = [path for path in _.allPaths if not pathHasGlobalIgnore(path, idirs) and not pathHasGlobalSkip(path, sdirs) and not anyParentIsSkipped(path, _.cfg.paths) and IGNORE not in dictGet(_.cfg.paths, path, {})]  # TODO add marker file logic below?
    if returnAll or len(include) == 0:  # if only exclusive tags, we need all paths and prune them later
      debug(f"Prune skipped and ignored paths from {len(_.allPaths)} to {len(alls)} paths")
      if returnAll:
        alls = [a for a in alls if os.path.isdir(_.root + os.sep + a)]  # ensure that only correct letter cases are retained on case-sensitive file systems
        return alls
    paths, first = (set() if len(include) else alls), True  # first filtering action (inclusive or exclusive)
    for tag in include:  # positive restrictive matching
      debug(f"Filter {len(alls if first else paths)} paths by inclusive tag <{tag}>")
      if isGlob(tag):  # filters indexed extensions by extension's glob (".c??"")
        new = reduce(lambda a, b: a.update(set(_.getPaths(_.tagdir2paths[_.tagdirs.index(b)], cache))) or a, normalizer.globfilter(_.tagdirs, tag), set())  # in-place version
      elif DOT in tag:
        new = wrapExc(lambda: set(_.getPaths(_.tagdir2paths[_.tagdirs.index(normalizer.filenorm(tag[tag.index(DOT):]))], cache)), set())  # directly from index
      else:  # no glob, no extension: tag or file name
        new = wrapExc(lambda: set(_.getPaths(_.tagdir2paths[_.tagdirs.index(tag)], cache)), set())  # directly from index, or a DOT-leading file/folder name
      if first: paths = new; first = False
      else: paths.intersection_update(new)
    for tag in exclude:  # we don't excluded globs here, because we would also exclude potential candidates (because index is over-specified)
      debug(f"Filter {len(paths)} paths by exclusive tag <{tag}>")
      potentialRemove = wrapExc(lambda: set(_.getPaths(_.tagdir2paths[_.tagdirs.index(tag)], cache)), set())  # these paths can only be removed, if no manual tag/file extension/glob in config or FROM. TODO what if positive extension looked for? already handled?
      new = _.removeIncluded(include, potentialRemove)  # remove paths with includes from "remove" list (adding back)
      if first:  # start with all paths, except determined excluded paths
        paths = _.allPaths - new
        first = False
      else:
        paths.difference_update(new)  # reduce found paths by exclude matches
    # Now convert to return list, which may differ from previously computed set of paths
    paths = [p for p in paths if not pathHasGlobalIgnore(p, idirs) and not pathHasGlobalSkip(p, sdirs)]  # TODO filter was done above already. remove filtering here, it should be reflected in index already (in contrast to above getall?)
    debug(f"Found {len(paths)} path matches")
    if _.cfg.case_sensitive:  # eliminate different letter cases, otherwise return both, although only one physically exists TOOD use "not"?
      paths[:] = [p for p in paths if os.path.isdir(_.root + os.sep + p)]  # ensure that only correct letter cases are retained on case-sensitive file systems
      debug(f"Retained {len(paths)} paths after removing duplicates")  # TODO on windows, this might succeed although case differs!
    assert all(path.startswith(SLASH) or path == '' for path in paths), paths  # only return root-relative paths
    return paths

  def findFiles(_, current, poss, negs):
    ''' Determine files for the given folder (from findFolders() with potential matchs).
        current: root-relative folder to filter files in
        poss:    (opt. case-normalized) positive (including) tags, file extensions, file names or globs
        negs:    negative (excluding) tags, file extensions, file names or globs
        returns: 2-tuple([filenames], skip?)
    '''
    debug(f"findFiles '{current}' {poss} {negs}")
    if current: assert current[0] == SLASH and current[-1] not in '/\\' and  _.root[-1] not in '/\\', f"{current} {_.root}"
    inPath = set(safeSplit(current, SLASH))  # break path into folder names
    inPath.update(reduce(lambda prev, step: prev + TOKENIZER.split(step), inPath, []))  # add tokenized path steps
    inPath.discard('')  # tokenizer may split an empty string
    inPath.update([normalizer.filenorm(f) for f in inPath])  # adds case-normalized versions if case_sensitive
    poss = set([p for p in poss if not normalizer.globfilter(inPath, p)])  # remove already true positive tags (folder name match)
    info(f"Filter folder '{current}' %s" % (("by remaining including tags <%s>" % (COMB.join(poss)) if len(poss) else (("by remaining excluding tags " + COMB.join(negs)) if len(negs) else "with no constraint"))))
    conf = _.cfg.paths.get(current, {})  # if empty we return all files
    mapped = [pathNorm(m if m.startswith(SLASH) else os.path.normpath(current + SLASH + m)) for m in conf.get(FROM, [])]  # root-absolute or folder-relative path
#    mapped[:] = [m for m in mapped if usInderRoot] TODO sanity check if path exists or if outside repository? could be considered a feature though
    if len(mapped): debug(f"Mapped folders: {os.pathsep.join(mapped)}")  # root-relative paths
    skipFilter = len(poss) + len(negs) + len(mapped) == 0

    willskip = False  # contains files from current or mapped folders (without path, since "mapped", but could get a local symlink - TODO
    found = set() if not skipFilter else set(wrapExc(lambda: [f for f in os.listdir(_.root + current) if isFile(_.root + current + os.sep + f)], []))  # TODO what if folder doesn't exist TODO duplicate of below as special case for no further constraints -> returns all
    if IGNFILE in found: skipFilter, found = True, set()  # enable skip but don't return anything
    for _f, folder in enumerate([] if skipFilter else [current] + mapped):
      info(("Check %s %%sfolder%%s" % ('mapped' if _f else 'proper')) % (('', f" '{folder}'") if folder else ("root ", "")))
      files = set(wrapExc(lambda: [f for f in os.listdir(_.root + folder) if isFile(_.root + folder + os.sep + f)], []))  # TODO silently catches for OS errors, e.g. encoding problems
      if IGNFILE in files: continue  # ignore is easy
      if folder == current and SKPFILE in files:  # only applies to non-mapped folder
        willskip = True  # skip is more difficult to handle than ignore, cf. return 2-tuple (call in tp.find())
        info(f"Skip '{folder}' due to local marker file")
        continue  # no break, process mapped folders
      caseMapping = dd()  # store mapping from case-normalized to actual filenames
      for f in files: caseMapping[normalizer.filenorm(f)].append(f)
      files.update(set(caseMapping.keys()))  # adds case-normalized file names for matching, used if case_sensitive == False

      tags = _.cfg.paths.get(folder, {}).get(TAG, [])

      keep = set(files)  # shallow copy. start with all files to keep, then reduce by remaining matching tag patterns
      for tag in poss:  # for every inclusive tag, check if action on current files is necessary
        if tag[0] == DOT:  keep.intersection_update(set(f for f in files if f[-len(tag):] == tag)); continue  # filter by file extension
        elif isGlob(tag):  keep.intersection_update(set(normalizer.globfilter(files, tag))); continue  # filter globs
        elif tag in files: keep.intersection_update(set([tag])); continue  # TODO cannot detect tokenized file globs here, we operate only on actual file system contents TODO what if same tag defined in remove

        diskeep = set()  # collect all disjunctive "keep" results for configured tag
        for value in (t for t in tags if t.startswith(tag + SEPA)):  # restrict by additional matching tags from the folder configuration
          assert value.count(SEPA) == 2
          tg, inc, exc = value.split(SEPA)  # tag name, includes, excludes
          conkeep = set(keep)  # all conjunctive filter expressions start with previously filtered set
          for i in safeSplit(inc):  # keep conjunction of inclusive files
            if i[0] == DOT: conkeep.intersection_update(set(f for f in conkeep if f[-len(i):] == i))  # TODO also check normalized extension here for user convenience? NO
            elif  i == ALL: continue  # keep all, nothing to do
            elif isGlob(i): conkeep.intersection_update(set(normalizer.globfilter(conkeep, i)))  # "set &"
            else:           conkeep.intersection_update(set([i]) if i in conkeep and isFile(_.root + folder + os.sep + i) else set())  # add file only if exists
          conrmve = set(keep) if exc else set()  # no excludes mean remove nothing TODO mirror this logic below?
          for e in safeSplit(exc):  # remove conjunction of exclusive files
            if e[0] == DOT: conrmve.intersection_update(set(f for f in conrmve if f[-len(e):] == e))
            elif isGlob(e): conrmve.intersection_update(set(normalizer.globfilter(conrmve, e)))
            else:           conrmve.intersection_update(set([i]) if i in conrmve and isFile(_.root + folder + os.sep + i) else set())
          diskeep.update(conkeep - conrmve)  # "set |" disjunctive combination of all filters for a tag
        keep.intersection_update(diskeep)
        if not keep: break  # no need to further check, if no matches remain after tag

      remove = set(files if negs else [])
      for tag in negs:
        if tag[0] == DOT:  remove.intersection_update(set(f for f in files if f[-len(tag):] == tag)); continue  # filter by file extension
        elif isGlob(tag):  remove.intersection_update(set(normalizer.globfilter(files, tag))); continue  # filter globs
        elif tag in files: remove.intersection_update(set([tag])); continue

        disremove = set()
        for value in (t for t in tags if t.startswith(tag + SEPA)):
          assert value.count(SEPA) == 2
          tg, inc, exc = value.split(SEPA)
          conremove = set(remove)  # collect all file that should be excluded
          for i in safeSplit(inc):
            if i[0] == DOT: conremove.intersection_update(set(f for f in conremove if normalizer.filenorm(f[-len(i):]) == normalizer.filenorm(i)))
            elif  i == ALL: continue
            elif isGlob(i): conremove.intersection_update(set(normalizer.globfilter(conremove, i)))
            else:           conremove.intersection_update(set([i]) if i in conremove and isFile(_.root + folder + os.sep + i) else set())
          conkeep = set(remove) if exc else set()
          for e in safeSplit(exc):
            if e[0] == DOT: conkeep.intersection_update(set(f for f in conkeep if normalizer.filenorm(f[-len(e):])))
            elif isGlob(e): conkeep.intersection_update(set(normalizer.globfilter(conkeep, e)))
            else:           conkeep.intersection_update(set([i]) if i in conkeep and isFile(_.root + folder + os.sep + i) else set())
          disremove.update(conremove - conkeep)
        remove.intersection_update(disremove)
        if not remove: break
      files.intersection_update(keep)
      files.difference_update(remove)
      debug(f"Files for folder '{folder}': %s (Keep/Remove: <%s>/<%s>)" % (COMB.join(files), COMB.join(keep - remove), COMB.join(remove)))
      found.update(set(reduce(lambda l, f: lappend(l, caseMapping.get(f, f)), files, set())))  # TODO may lead to different cases or shadowing if using mapped dirs
    debug(f"findFiles '{current}' returns {list(found)} skip: {willskip}")  # TODO in contrast to findFolder no file exist checks (mapped entries are harder to check). ? only partially with config TAGS
    return found, willskip


if __name__ == '__main__': import doctest; doctest.testmod()



# coding=utf-8

''' tagsPlorer command-line interface  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

import logging, optparse, os, sys, time
assert sys.version_info >= (3, 6), "tagsPlorer requires Python 3.6+"

from tagsplorer.constants import ALL, APPNAME, COMB, CONFIG, DOT, FROM, GLOBAL, INDEX, RIGHTS, SLASH, ST_MTIME
from tagsplorer.lib import Configuration, Indexer
from tagsplorer.utils import caseCompareKey, casefilter, dd, dictGetSet, isDir, isGlob, isUnderRoot, lindex, normalizer, pathNorm, removeTagPrefixes, safeSplit, safeRSplit, sjoin, splitByPredicate, splitTags, wrapExc, xany
from tagsplorer import lib, simfs, utils  # for setting the log level dynamically


logging.basicConfig(
  level =  logging.DEBUG if '-V' in sys.argv or '--debug'   in sys.argv or os.environ.get("DEBUG",   "False").lower() == "true" else
          (logging.INFO  if '-v' in sys.argv or '--verbose' in sys.argv or os.environ.get("VERBOSE", "False").lower() == "true" else
           logging.WARNING),  # must be set here to already limit writing to info and debug (e.g. "Started at...")
  stream = sys.stdout if '--stdout' in sys.argv else sys.stderr,  # log to stderr, write results to stdout
  format =  '%(asctime)-8s.%(msecs)03d %(levelname)-4s %(module)s:%(funcName)s:%(lineno)d | %(message)s',
  datefmt = '%H:%M:%S')
wrapExc(lambda: sys.argv.remove('--stdout'))  # remove if present
_log = logging.getLogger(__name__)
def log(func): return (lambda *s: func(sjoin([_() if callable(_) else _ for _ in s]), **({"stacklevel": 2} if sys.version_info >= (3, 8) else {})))
debug, info, warn, error = log(_log.debug), log(_log.info), log(_log.warning), log(_log.error)


with open(os.path.join(os.path.dirname(__file__), 'VERSION'), encoding = 'utf-8') as fd: VERSION = fd.read()
APPSTR  = APPNAME + " version %s  (C) 2016-2021  Arne Bachmann" % VERSION


def findRootFolder(filename, start = None):
  ''' Utility function that tries to auto-detect a filename in any parent folder.
      filename: file to find
      returns: existing file path or None if not found
  >>> findRootFolder(os.path.basename(os.path.abspath(__file__)), start = os.path.join(os.path.abspath(__file__), 'a', 'b')) == os.path.abspath(__file__)
  True
  '''
  path = start if start else os.path.abspath(os.getcwd())
  try:
    while not os.path.exists(os.path.join(path, filename)):
      new = os.path.dirname(path)
      if new == path: return None  # avoid infinite loop when hitting root
      path = new
    return os.path.join(path, filename)
  except Exception as E: error(E)


def getRoot(options, args):
  ''' Determine tagsPlorer repository root folder by checking program arguments with some fall-back logic.
      returns 2-tuple(root folder absolute path, index/config file folder absolute path), depending on options
  >>> class O:
  ...   def __init__(_): _.index = _.root = None  # options object
  >>> os.chdir(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, "_test-data"))  # change into folder that contains .tagsplorer.cfg
  >>> o = O(); rela = lambda a, b: (os.path.relpath(a, start = os.getcwd()), os.path.relpath(b, start = os.getcwd()))  # helper to pass with any current folder
  >>> rela(*getRoot(o, []))  # no options nor arguments: root folder == index folder
  ('.', '.')
  >>> o.root = "./x"; rela(*getRoot(o, []))  # given a root: root = index
  ('x', 'x')
  >>> o.index = "./y"; rela(*getRoot(o, []))  # given both
  ('x', 'y')
  >>> o.root = None; o.index = "abc"; rela(*getRoot(o, []))
  ('abc', 'abc')
  '''
  if options.index:  # given folder to place the index
    if options.root is None: options.root = str(options.index)  # copy by value
    return (os.path.abspath(options.root), os.path.abspath(options.index))
  # determine root folder by finding the config/index files
  folder = options.root  # None if not defined
  if folder is None: folder = wrapExc(lambda: os.path.relpath(os.path.dirname(findRootFolder(CONFIG))))
  if folder is None: folder = os.curdir  # fallback
  root, meta = os.path.abspath(folder), os.path.abspath(options.index if options.index else folder)
  debug(f"Root and index folders found at {root} {meta}")
  return root, meta


class CatchExclusionsParser(optparse.OptionParser):
  ''' Allows to process undefined options as non-option arguments (e.g. --tag as an additional exclusive tag marked via --). '''
  def __init__(_):
    ihf = optparse.IndentedHelpFormatter(2, 60, 120)  # , formatter = optparse.TitledHelpFormatter(),
    optparse.OptionParser.__init__(_, add_help_option = False, prog = APPNAME, usage = "python tp <tags and options>", description = APPSTR, version = VERSION, formatter = ihf)

  def _process_args(_, largs, rargs, values):
    while rargs:
      try: optparse.OptionParser._process_args(_, largs, rargs, values)
      except (optparse.BadOptionError, optparse.AmbiguousOptionError) as E: largs.append(E.opt_str)


class Main:
  ''' This is the main business logic, operated by the command-line commands. '''

  def initIndex(_):
    ''' Set up and store an empty index, thus marking the repository root folder. '''
    if not _.options.root:  _.options.root  = os.getcwd()  # if not specified on command line
    if not _.options.index: _.options.index = os.getcwd()
    folder, meta = getRoot(_.options, _.args)
    if _.options.strict and os.path.exists(os.path.join(meta, CONFIG)): error("Index already exists. Use --force or -f to override"); return 1
    info(f"Create root configuration at '{meta}'")
    cfg = Configuration()
    if not _.options.simulate:
      try: os.makedirs(meta, mode = RIGHTS, exist_ok = True)
      except OSError as E: debug(f"Cannot create folder {E}")
      if not os.path.exists(meta): error("Could not create index for that folder. Check access rights and path errors"); return 2
      cfg.store(meta)
    return 0

  def updateIndex(_):
    ''' Crawl the folder tree and (re)create the index file.
        _.options: options structure from optparse, set in parse
        _.args:    arguments list from optparse, set in parse
        returns: 2-tuple(created Index, exit code if error)
    '''
    folder, meta = getRoot(_.options, _.args)
    cfg = Configuration()
    cfg.load(meta)
    stop = False
    for path, defs in cfg.paths.items():  # check for correctness
      for other in defs.get(FROM, []):
        abspath = (folder + other) if other.startswith(SLASH) else os.path.normpath(folder + path + SLASH + other)
        if   not isUnderRoot(folder, abspath): error(f"Configured mapped folder '{other}' for '{path}' is outside indexed folder tree, please fix"); stop = True
        elif not isDir(              abspath): error(f"Configured mapped folder '{other}' for '{path}' not found, please fix"); stop = True
    if stop: return None, 1
    idx = Indexer(folder)  # no need to load the old index
    idx.walk(cfg)  # track all files using the configuration settings
    if not _.options.simulate: idx.store(os.path.join(meta, INDEX), idx.timestamp)
    return idx, 0

  def find(_):
    ''' Find and display all folders that match the provided tags in _.options.includes while excluding those from _.options.excludes.
        returns: exit code
    '''
    poss, negs = splitByPredicate(splitTags(_.args), lambda e: e[0] != '-')  # allows direct arguments, potentially suffixed by +/-
    poss.extend(splitTags(_.options.includes))
    negs.extend(splitTags(_.options.excludes))
    poss, negs = removeTagPrefixes(poss, negs)

    folder, meta = getRoot(_.options, _.args)
    indexFile = os.path.join(meta, INDEX)
    if not os.path.exists(indexFile):  # e.g. first run after root initialization
      error("No index file found. " + ("Exit" if _.options.keep_index else "Crawl folder tree"))
      if _.options.keep_index: return 2
      idx, code = _.updateIndex()  # crawl folder tree immediately and return search index
      if code: return code
    else:
      idx = Indexer(meta)
      idx.load(indexFile, ignore_skew = _.options.keep_index)  # load search index from root
    normalizer.setupCasematching(not (_.options.ignore_case or not idx.cfg.case_sensitive))  # case option can be overriden by --ignore-case
    poss, negs = map(lambda l: list(map(normalizer.filenorm, l)), (poss, negs))  # convert search terms to normalized case, if necessary
    debug("Effective search filters +<%s> -<%s>"        % (COMB.join(poss), COMB.join(negs)))
    _exts = [ext for ext in poss + negs if ext and ext[0] == DOT]
    if len(_exts) > 1:
      error("Cannot match anything if more than one file extension is specified (%s)" % COMB.join(_exts)); return 1

    info(f"Search '{idx.root}' for tags +<%s> -<%s>" % (COMB.join(poss), COMB.join(negs)))
    paths = idx.findFolders(poss, negs)
    debug(f"Found {len(paths)} potential path matches")
    #[debug(path) for path in paths]  # optimistic: all folders "seem" to match given tags, but only some might actually do (e.g. due to excludes)

    if len(paths) == 0 and xany(lambda x: isGlob(x) or DOT in x, poss + negs):  # glob or extension
      paths = idx.findFolders([], [], True)  # return all folders names unfiltered (except ignore/skip without marker files)
    if _.options.onlyfolders:
#      for p in poss: paths[:] = [x for x in paths if not isGlob(p) or     normalizer.globmatch(safeRSplit(x), p)]  # successively reduce paths down to matching positive tags: in --dirs mode tags currently have to be folder names TODO later we should reflect actual mapping
#      for n in negs: paths[:] = [x for x in paths if not isGlob(n) or not normalizer.globmatch(safeRSplit(x), n)]  # TODO is this too strict and ignores configured tags and the index entirely? TODO also handle tokenization here -> not necessary, as already found in findFolders since tokenized in the index
      info(f"Found {len(paths)} folders for +<%s> -<%s>" % (COMB.join(poss), COMB.join(negs)))
      try:
        if len(paths): print("\n".join((idx.root if not _.options.relative else '') + path for path in paths))
      except KeyboardInterrupt: pass  # idx.root + path + SLASH + file for file in files)); counter += len(files)
      return 0  # no file filtering requested

    dcount, counter, skipped, run = 0, 0, [], None  # if showing also files
    for path, fskip in ((p, idx.findFiles(p, poss, negs)) for p in paths):
      files, skip = fskip  # deconstruct 2-tuple
      if skip: skipped.append(path); continue  # memorize prefix to skip all folders under it HINT relies on breadth-first traversal (since ord('.') < ord('/') < ord('A'))
      if xany(lambda skp: path.startswith(skp), skipped): continue  # is in skipped folder tree HINT if root is skipped, will if course ignore all subfolders as well
      dcount += 1
      try:
        if len(files) > 0:
          print("\n".join((idx.root if not _.options.relative else '') + path + SLASH + file for file in files))
          counter += len(files)  # incremental output
          run = list(files)[0]
      except KeyboardInterrupt: break
    info(f"Found {counter} files in {dcount} folders for +<%s> -<%s>" % (COMB.join(poss), COMB.join(negs)))
    if counter == 1 and _.options.run:
      import subprocess
      with subprocess.Popen(f'''{'start "tagsPlorer run"' if sys.platform == "win32" else ""} "{run}"''', shell = True) as process:
        try: sys.exit(process.wait(5))
        except KeyboardInterrupt: process.kill(); return 1
        except subprocess.TimeoutExpired: warn("Still running")
    return 0

  def config(_, unset = False, get = False, all = False):
    ''' Define, display or remove a global configuration parameter. '''
    value = ((_.options.setconfig if not get else (_.options.getconfig if not all else None)) if not unset else _.options.unsetconfig)
    if value is None and not all: warn("No configuration key provided"); return 1
    if not unset and not get and "=" not in value: warn("Configuration entry must be specified in the form <key>=<value>"); return 2
    key, value = safeSplit(value, "=")[:2] if not unset and not get else (value, None)
    key = wrapExc(lambda: key.lower())  # config keys are normalized to lower case
    folder, meta = getRoot(_.options, _.args)
    cfg = Configuration()
    cfg.load(meta)
    if get:  # get operation
      if all:
        for k, v in ((_k, cfg.__dict__[_k]) for _k in cfg.__dict__ if _k != "paths"): warn(f"Configuration entry: {k} = {v}")
        return 0
      elif key in cfg.__dict__: warn(f"Configuration entry: {key} = {cfg.__dict__[key]}"); return 0
      else: warn(f"Configuration key '{key}' not found"); return 3
    if '' not in cfg.paths: cfg.paths[''] = dd()  # create global section
    entries = dictGetSet(cfg.paths[''], GLOBAL, [])
    index = wrapExc(lambda: lindex([kv.split("=")[0].lower() for kv in entries], key))  # find index for specified key, or None
    if not unset:  # set operation
      if index is not None: entries[index] = f"{key}={value}"; warn(f"Updated configuration entry: {key} = {value}")
      else: entries.append(f"{key}={value}"); warn(f"Added configuration entry: {key} = {value}")
      cfg.__dict__[key] = value if value.lower() not in ("true", "false") else value.lower() == "true"
    else:  # unset operation
      if index is not None: del entries[index]; warn("Removed configuration entry for key '%s'" % key)
      else: warn("Configuration entry not found, nothing to do")
      try: del cfg.__dict__[key]  # remove in-memory global configuration
      except: pass
    if not _.options.simulate: cfg.store(meta)
    return 0

  def reset(_):
    ''' Remove all global configuration entries. '''
    folder, meta = getRoot(_.options, _.args)
    cfg = Configuration()
    cfg.load(meta)
    cfg.reset()  # to defaults
    if not _.options.simulate: cfg.store(meta)
    warn("Reset configuration parameters")
    return 0


  def assign(_):
    ''' Add one or more tags for the given (inclusive and/or exclusive) globs.
        TODO detect if all existing patterns are mutually exclusive or probably always covered by other globs
    '''
    folder, meta = getRoot(_.options, _.args)
    root = pathNorm(folder)
    cfg = Configuration()
    cfg.load(meta)

    tags = safeSplit(_.options.tag)  # from -t option
    poss, negs = splitByPredicate(tags, lambda e: e[0] != '-')  # TODO only additive tags allowed (index is over-specified and removing is hard?)
    poss, negs = removeTagPrefixes(poss, negs)
    if (len(poss) + len(negs)) == 0: error("No tag(s) provided to assign for " + COMB.join(_.args)); return 1
    if xany(lambda p: p in negs, poss): error("Same tag in both inclusive and exclusive pattern " + ', '.join(["'%s'" % p for p in poss if p in negs])); return 2

    args = [pathNorm(arg) for arg in _.args]
    modified = False
    for pathpatterns in args:
      path, patterns = (pathpatterns[:pathpatterns.rindex(SLASH)], pathpatterns[pathpatterns.rindex(SLASH) + 1:]) if SLASH in pathpatterns else (os.getcwd(), pathpatterns)
      inc, exc = splitByPredicate(patterns.split(COMB), lambda e: e[0] != '-')
      inc, exc = removeTagPrefixes(inc, exc)
      path = pathNorm(os.path.abspath(os.path.normpath(path)))
      if not isUnderRoot(root, path):
        error(f"Specified pattern path '{path}' is outside indexed folder tree '{root}'; skip '{pathpatterns}'"); continue
      if not os.path.exists(path):
        error(f"Specified pattern path '{path}' does not exist; skip '{pathpatterns}'"); continue
      path = path[len(root):]  # make repo-relative
      for pn in (poss + negs):
        if pn in safeSplit(path, SLASH): warn(f"Specified tag <{pn}> should not match any path constituents of '{path}'"); continue  # TODO unless positive, and folder is ignored somehow
      assert path[0] == SLASH
      files = os.listdir(root + path)  # check current state on file system
      _i, _e = [], []
      for ps, l, t in zip((inc, exc), (_i, _e), ("Inclusive", "Exclusive")):  # performs sanity checks on inclusive and exclusive patterns
        for p in ps:
          exists = False
          if   p == '': error(f"{t} tag pattern for '{path}' cannot be empty, skip"); continue
          if p[0] == DOT: exists = len(casefilter(files, p)) > 0
          elif  p == ALL: exists = len(files) > 0
          elif isGlob(p): exists = len(casefilter(files, p)) > 0
          else:           exists = p in files
          if not exists:
            warn(f"No file matches for glob '{path}/{p}'%s'" % (", skip" if _.options.strict else ", add anyway"))
            if _.options.strict: continue
          l.append(p)
      if len(_i + _e):
        for tag in poss: modified = cfg.addTag(path, tag, _i, _e, not _.options.strict) or modified  # TODO also add negative tags, if semantics clear
    if not modified: error("Nothing was added")
    if modified and not _.options.simulate: cfg.store(meta)
    return 0

  def remove(_):
    ''' Remove one or more tags for the given (inclusive and/or exclusive) globs. '''
    folder, meta = getRoot(_.options, _.args)
    root = pathNorm(folder)
    cfg = Configuration()
    cfg.load(meta)

    tags = safeSplit(_.options.untag)  # from -u option
    poss, negs = splitByPredicate(tags, lambda e: e[0] != '-')
    poss, negs = removeTagPrefixes(poss, negs)
    if (len(poss) + len(negs)) == 0: error("No tag(s) provided to remove for %s" % COMB.join(_.args)); return 1
    if xany(lambda p: p in negs, poss): error("Same tag in both inclusive and exclusive pattern " + ', '.join(["'%s'" % p for p in poss if p in negs])); return 2

    args = [pathNorm(arg) for arg in _.args]
    modified = False
    for pathpatterns in args:
      path, patterns = (pathpatterns[:pathpatterns.rindex(SLASH)], pathpatterns[pathpatterns.rindex(SLASH) + 1:]) if SLASH in pathpatterns else (os.getcwd(), pathpatterns)
      inc, exc = splitByPredicate(patterns.split(COMB), lambda e: e[0] != '-')
      inc, exc = removeTagPrefixes(inc, exc)
      path = pathNorm(os.path.abspath(os.path.normpath(path)))
      if not isUnderRoot(root, path):
        error(f"Specified pattern path '{path}' is outside indexed folder tree '{root}'; skip '{pathpatterns}'"); continue
      path = path[len(root):]  # make repo-relative
      assert path[0] == SLASH
      files = os.listdir(root + path)  # check current state on file system
      _i, _e = [], []
      for ps, l, t in zip((inc, exc), (_i, _e), ("Inclusive", "Exclusive")):  # performs sanity checks on inclusive and exclusive patterns
        for p in ps:
          exists = False
          if   p == '': error(f"{t} tag pattern for '{path}' cannot be empty, skip"); continue
          if p[0] == DOT: exists = len(casefilter(files, p)) > 0
          elif  p == ALL: exists = len(files) > 0
          elif isGlob(p): exists = len(casefilter(files, p)) > 0
          else:           exists = p in files
          if not exists:
            warn(f"No file matches for glob '{path}/{p}'%s'" % (", skip" if _.options.strict else ", remove anyway"))
            if _.options.strict: continue
          l.append(p)

      if len(_i + _e):
        for tag in poss: modified = cfg.delTag(path, tag, _i, _e) or modified
    if not modified: error("Nothing was removed")
    if modified and not _.options.simulate: cfg.store(meta)
    return 0

  def show(_):
    ''' Show tags for current or specified folder. '''
    folder, meta = getRoot(_.options, _.args)
    root = pathNorm(folder)
    cfg = Configuration()  # load config early for the case_sensitive flag
    cfg.load(meta)

    for folder in ([os.getcwd()] if not _.args else _.args):
      warn(f"Tags for {folder}")
      folder = pathNorm(os.path.abspath(folder))
      folder = folder.rstrip(SLASH + os.sep)
      if not isUnderRoot(root, folder): warn(f"Relative path '{folder}' is outside indexed folder tree '{root}'; skip"); continue
      cfg.showTags(folder[len(root):])


  def stats(_):
    ''' Show index stats, optionally including very detailed information.
        returns: exit code
    '''
    folder, meta = getRoot(_.options, _.args)
    indexFile = os.path.join(meta, INDEX)
    if not os.path.exists(indexFile):
      error("No index file found. %s" % ("Exit" if _.options.keep_index else "Crawl folder tree"))
      if _.options.keep_index: return 2
      idx, code = _.updateIndex()  # crawl folder tree immediately and return search index
      if code: return code
    else:
      idx = Indexer(meta)
      idx.load(indexFile, ignore_skew = _.options.keep_index)
    _.options.relative = True  # don't output full paths here
    warn("Configuration stats:")
    warn("  Compression level:", idx.cfg.compression)
    warn("  Number of configured paths: %d" % len(idx.cfg.paths))
    warn("  Average number of markers per folder: %.2f" % ((sum([len(_) for _ in idx.cfg.paths.values()]) / len(idx.cfg.paths)) if idx.cfg.paths else 0.))  # e.g. skip, ignore, manual tags
    warn("  Average number of entries per folder: %.2f" % ((sum([sum([len(__) for __ in _.values()]) for _ in idx.cfg.paths.values()]) / len(idx.cfg.paths)) if idx.cfg.paths else 0.))  # e.g. skip
    warn("  Last update: " + time.strftime("%Y-%m-%d@%H:%M", time.localtime(os.stat(os.path.join(meta, CONFIG))[ST_MTIME])))
    warn("Index stats:")
    warn("  Root folder:", idx.root)
    warn("  Timestamp:", time.strftime("%Y-%m-%d@%H:%M", time.localtime(idx.timestamp / 1000.)))
    warn("  Timestamp (ms epoch):", idx.timestamp)
    warn("  Number of tags:",   len(idx.tagdirs))
    info("Tags and folders:")  # (occurrence = same name for different folder name references")
    if not _.options.verbose and not _.options.debug_on: return 0
    byOccurrence = dd()
    for i, t in enumerate(frozenset(idx.tagdirs)):
      byOccurrence[idx.tagdirs.count(t)].append(i)  # map number of tag occurrences in index to their tagdir indices
    _cache = {}
    for n, ts in sorted(byOccurrence.items()):
      info(f"  {n} occurence%s for entries %s" % ("s" if n > 1 else "", COMB.join([str(_) for _ in sorted(ts)])))
      if not _.options.debug_on: return 0
      byMapping = dd()
      for t in ts: byMapping[idx.tagdirs[t]].extend(idx.tagdir2paths[t])  # aggregate all mappings
      for t in sorted(byMapping.keys(), key = caseCompareKey):
        debug(f"    Entry '{t}' (%s) maps to: %s" % (COMB.join([str(i) for i, x in enumerate(idx.tagdirs) if x == t]), ', '.join(["%s (%d)" % (idx.getPath(_i, _cache), _i) for _i in byMapping[t]])))
    return 0

  def parse_and_run(_):
    ''' Main logic that analyses the command line arguments and starts an operation. '''
    ts = time.time()
    info("Started at %s" % (time.strftime("%H:%M:%S", time.localtime(ts))))
    op = CatchExclusionsParser()  # https://docs.python.org/3/library/optparse.html#optparse-option-callbacks HINT options default to None!
    op.add_option('-I', '--init',           action = "store_true",  dest = "init",        default = False,             help = "Create empty index (repository root)")
    op.add_option('-r', '--root',           action = "store",       dest = "root",        default = None,  type = str, help = "Specify root folder of file tree, default: current folder")
    op.add_option('-i', '--index',          action = "store",       dest = "index",       default = None,  type = str, help = "Specify alternative index folder (if different from root)")
    op.add_option('-U', '--update',         action = "store_true",  dest = "update",      default = False,             help = "Force-update the index, crawl files in folder tree")
    op.add_option('-s', '--search',         action = "append",      dest = "includes",    default = [],                help = "Find files by tags (default action if no option specified)")
    op.add_option('-x', '--exclude',        action = "append",      dest = "excludes",    default = [],                help = "Tags to ignore. Same as -<tag>")
    op.add_option('-t', '--tag',            action = "store",       dest = "tag",         default = None,  type = str, help = "Set   tag(s) for given file(s) or glob(s): tp -t tag,tag2,-tag3... file,glob...")
    op.add_option('-T', '--untag',          action = "store",       dest = "untag",       default = None,  type = str, help = "Unset tag(s) for given file(s) or glob(s): tp -d tag,tag2,-tag3... file,glob...")
    op.add_option(      '--tags',           action = "store_true",  dest = "show_tags",   default = False,             help = "List defined tags for a folder")
    op.add_option(      '--get',            action = "store",       dest = "getconfig",   default = None,  type = str, help = "Get a global configuration parameter [<key>]")
    op.add_option(      '--set',            action = "store",       dest = "setconfig",   default = None,  type = str, help = "Set a global configuration parameter <key>=<value>")
    op.add_option(      '--unset',          action = "store",       dest = "unsetconfig", default = None,  type = str, help = "Unset a global configuration parameters")
    op.add_option(      '--config',         action = "store_true",  dest = "showconfig",  default = None,              help = "Display all global configuration parameters")
    op.add_option('-R', '--reset',          action = "store_true",  dest = "resetconfig", default = False,             help = "Reset all global configuration parameters to defaults")
    op.add_option(      '--run',            action = "store_true",  dest = "run",         default = False,             help = "Attempt to run file, if search results in exactly one match")
    op.add_option('-f', '--force',          action = "store_false", dest = "strict",      default = True,              help = "Force operation, relax safety measures")  # mapped to inverse "strict" flag
    op.add_option('-c', '--ignore-case',    action = "store_true",  dest = "ignore_case", default = False,             help = "Search case-insensitive (overrides option in index)")
    op.add_option('-n', '--simulate',       action = "store_true",  dest = "simulate",    default = False,             help = "Don't write anything")
    op.add_option('-k', '--keep-index',     action = "store_true",  dest = "keep_index",  default = False,             help = "Don't update the index, even if configuration was changed")
    op.add_option(      '--dirs',           action = "store_true",  dest = "onlyfolders", default = False,             help = "Only find folders that contain matches")
    op.add_option('-v', '--verbose',        action = "store_true",  dest = "verbose",     default = False,             help = "Display more information")
    op.add_option('-V', '--debug',          action = "store_true",  dest = "debug_on",    default = False,             help = "Display internal data state")
    op.add_option(      '--stats',          action = "store_true",  dest = "stats",       default = False,             help = "List index internals")
    op.add_option(      '--relative',       action = "store_true",  dest = "relative",    default = False,             help = "Output files with root-relative paths only")  # instead of absolute file system paths
    op.add_option(      '--simulate-winfs', action = "store_true",  dest = "winfs",       default = True,              help = optparse.SUPPRESS_HELP)  # "Simulate case-insensitive file system")  # but option is checked outside parser in simfs.py
    op.add_option('-h', '--help',           action = "help",                                                           help = optparse.SUPPRESS_HELP)  # whoever showed the help, doesn't need this information
    _.options, _.args = op.parse_args()  # TODO replace with argparse?
    reserved1, reserved2 = (set(_) for _ in splitByPredicate([_.get_opt_string() for _ in op.option_list], lambda e: e[:2] != '--'))  # handle reserved option switches that must be masked by a dash to operate as an exclude tag
    _.args, excludes = splitByPredicate(_.args,   lambda e: e[:2] != '---')      # split definitive excludes (triple dash)
    _.args, exclude_ = splitByPredicate(_.args,   lambda e: e[:2] != '--')       # split potential  excludes (double dash)
    add2, exclude2   = splitByPredicate(exclude_, lambda e: e not in reserved2)  # filter out program options with double dash
    add1, exclude1   = splitByPredicate(exclude2, lambda e: e not in reserved1)  # filter out program options with single dash
    _.options.excludes.extend([_.strip("---") for _ in excludes] + [_.strip("--") for _ in add2] + [_.strip("-") for _ in add1])  # update excludes option
    logLevel = logging.DEBUG if _.options.debug_on else (logging.INFO if _.options.verbose else logging.WARNING)
    _log.setLevel(logLevel)
    for mod in (lib, simfs, utils): mod._log.setLevel(logLevel)
    debug(f"Options:   {_.options}")
    debug(f"Arguments: {_.args}")
    code = 0  # exit code
    if   _.options.init:        code = _.initIndex()
    elif _.options.update: idx, code = _.updateIndex()
    elif _.options.tag:         code = _.assign()
    elif _.options.untag:       code = _.remove()
    elif _.options.show_tags:          _.show()
    elif _.options.setconfig:   code = _.config()
    elif _.options.unsetconfig: code = _.config(unset = True)
    elif _.options.getconfig:   code = _.config(get   = True)
    elif _.options.showconfig:  code = _.config(get   = True, all = True)
    elif _.options.resetconfig: code = _.reset()
    elif _.options.stats:       code = _.stats()
    elif _.args \
      or _.options.includes \
      or _.options.excludes:    code = _.find()
    else: error("No option specified. Use '--help' to list all options")
    info("Finished at %s after %.1fs" % (time.strftime("%H:%M:%S"), time.time() - ts))
    sys.exit(code)


def main(): Main().parse_and_run()  # Main entry point for console tools (setuptools)


if __name__ == '__main__':   # pragma: no cover
  if '--test' in sys.argv: import doctest; sys.exit(doctest.testmod(optionflags = doctest.ELLIPSIS)[0])
  main()



# TODO pipe stdin to allow manual reindex and svn updates
# report number of cache entries before each cache cleaning to graphite

import os, subprocess, sys, time
import logging; logging.basicConfig(level = logging.DEBUG if os.environ['DEBUG'].lower() == 'true' or '--debug' in sys.argv[:2] else logging.INFO, stream = sys.stdout, format = "%(asctime)-25s %(levelname)-8s %(name)s | %(message)s"); from logging import debug, info, warn, error

firstRepo = 2 if '--debug' in sys.argv[:2] else 1
REPOS, METAS = zip(*[tuple(_.split(",")) for _ in sys.argv[firstRepo:]] if len(sys.argv) > firstRepo else (os.getcwd(), os.getcwd()))  # indexed folder trees to search
PYTHON = os.path.realpath(sys.executable) if sys.platform != 'win32' else '"' + os.path.realpath(sys.executable) + '"'
CACHE = os.path.join(os.getcwd(), "http", "cache")

def call(command): return subprocess.Popen(command, shell = True, bufsize = 1, stdout = sys.stdout, stderr = sys.stderr)

def clearCache():
  for file in os.listdir(CACHE):
    try: os.unlink(file)
    except: logging.warn("Cannot remove cache entry " + repr(file))

logging.info("Starting tagsPlorer HTTP server supervisor")
clearCache()
while True:  # infinite loop
  # First thing each night (or at first start) is to update the source, then re-create all indexes
  logging.info("Updating tagsPlorer source code base")  # but keeps custom files present and merges local modifications
  if call("svn up").wait() != 0: logging.error("Error updating source code base (svn missing?), continuing with previous code base")
  for repo, meta in zip(REPOS, METAS):
    logging.info("Indexing repository '%s' from configuration '%s'" % (repo, meta))
    if call(PYTHON + " tp.py -r %s -i %s -uv" % (repo, meta)).wait() != 0: logging.error("Error indexing repository %s" % repo)
  logging.info("Starting tagsPlorer HTTP server process(es)")
  p = call(PYTHON + " serve.py " + " ".join([r + "," + m for (r, m) in zip(REPOS, METAS)]))  # TODO allow sending SIGINT from the supervisor (e.g. for nightly restart)
  while True:
    time.sleep(300.)  # 5 minutes waiting
    p.poll()
    # TODO check state: if exited - break
    # TODO check hour of day, clearCache() an recreate index every night, break
logging.info("Shut down tagsPlorer HTTP server supervisor")



''' tagsPlorer library  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

import os, re, sys


# Constants
MAJOR_VERSION = 0
APPNAME = "tagsPlorer"  # or something clunky like "virtdirview"
RIGHTS  = 0o760  # for creating new index folders (usually exist already)
CONFIG  = ".tagsplorer.cfg"  # main user-edited configuration file
INDEX   = ".tagsplorer.idx"  # index         file (re-built on every manual or detected file system change)
SKPFILE = ".tagsplorer.skp"  # skip   marker file (could equally be configured in configuration instead)
IGNFILE = ".tagsplorer.ign"  # ignore marker file (could equally be configured in configuration instead)
IGNORE, SKIP, TAG, FROM, SKIPD, IGNORED, GLOBAL = "ignore", "skip", "tag", "from", "skipd", "ignored", "global"  # allowed config file options
COMB, SEPA, SLASH, DOT, ALL, ST_MTIME, ST_SIZE = ",", ";", "/", os.extsep, "*", 8, 6  # often-used constants
TOKENIZER = re.compile(r"[\s\-_\.]+")  # tokenize file names as additional tags WARN may return empty string
REGEX_SANITIZE = re.compile(r"([\.\[\]\(\)\^\$\+\*\?\{\}])")  # all special regex characters to escape
PICKLE_PROTOCOL = 4  # (Python V3.4+) for pypy3 compatibility
SKIPDS = ["$RECYCLE.BIN", "System Volume Information"]
IGNOREDS = []


ON_WINDOWS = sys.platform == 'win32'



''' tagsPlorer file system emulation  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

import doctest, logging, os, sys, unittest


_log = logging.getLogger(__name__)


if '--simulate-winfs' in sys.argv or os.environ.get("SIMULATE_WINFS", "false").lower() == "true":
  _log.debug("Using case-insensitive file system emulation (for testing Windows behavior on Linux)")
  if '--simulate-winfs' in sys.argv: sys.argv.remove('--simulate-winfs')

  from tagsplorer.constants import SLASH, ST_SIZE
  from tagsplorer.utils import wrapExc

  SIMFS = True
  _RIGHTS = 0o777

  def _realPath(path):
    ''' Central function that determines the actual file name case for the given folder or file path.
        path:    file system path to analyse
        returns: string for real path as stored on the file system
    >>> with open("_x.X", "w"): pass  # create file
    >>> print(os.path.exists("_x.X"))
    True
    >>> with open("_x.X", "a"): pass  # write same
    >>> print(os.path.exists("_x.X"))
    True
    >>> fd = open("_x.X", "w"); fd.write("123"); fd.close()
    3
    >>> print(os.path.exists("_x.X"))
    True
    >>> fd = open("_X.x", "w"); fd.write("234"); fd.close()  # write over different case of same file using file handle
    3
    >>> print(os.path.exists("_X.x"))
    True
    >>> print(os.path.exists("_x.X"))
    True
    >>> with open("_X.X", "a") as fd: fd.write("567")
    3
    >>> fd = open("_x.x", "r"); contents = fd.read(); fd.close(); print(contents)
    234567
    >>> wrapExc(lambda: _unlink("_x.X"), None)
    >>> print(os.path.exists("_X.x"))
    False
    >>> print(os.path.exists("_x.X"))
    False
    '''
    _log.debug(f"_realPath '{path}'")
    assert path is not None
    if not path: return path
  #  path = _normpath(path)  # remove intermediate ".." etc. TODO leads to recursion due to use of os.stat
    steps = path.replace(os.sep, SLASH).split(SLASH)
    absolute = path[0] == SLASH
    if steps[0] not in ('', os.curdir): steps.insert(0, os.curdir)  # must be relative path like a/b, (otherwise ./a/b or /a/b, but never A:/b/c, because not on Windows)
    while os.pardir in steps:  # implement os.path.normpath
      steps.pop(steps.index(os.pardir) - 1)  # /a/b/../c = /a/c
      steps.remove(os.pardir)  # /a/b/../c = /a/c
    while os.curdir in steps: steps.remove(os.curdir)  # implement os.path.normpath
    if absolute: steps.pop(0)
    real = '' if absolute else '.'
    for step in steps:
      try: files = _listdir(real if real else '/')
      except: real += SLASH + step; continue  # cannot access: continue with path as given
      found = [f for f in files if f.lower() == step.lower()]  # match case-normalized
      if len(found) != 1:
        if len(found) > 1: debug(f"Cannot determine real path, since multiple names match {step}: {found}"); found = list(reversed(sorted(found)))[:1]
        else: found = [step]  # fallback
      real += SLASH + found[0]
    return real


  _exists = os.path.exists
  def __exists(path): return wrapExc(lambda: _exists(_realPath(path)), False)  # also for file handles
  os.path.exists = __exists  # monkey-patch function

  _unlink, _remove = os.unlink, os.remove  # could be the same, but just in case
  def __unlink(path): return _unlink(_realPath(path))
  def __remove(path): return _remove(_realPath(path))
  os.unlink, os.remove = __unlink, __remove

  _normpath = os.path.normpath
  def __normpath(path): return _normpath(_realPath(path))  # in Coconut: def os.path.isdir = ...
  os.path.normpath = __normpath

  _isdir = os.path.isdir
  def __isdir(path): return _isdir(_realPath(path))  # in Coconut: def os.path.isdir = ...
  os.path.isdir = __isdir

  _isfile = os.path.isfile
  def __isfile(path): return _isfile(_realPath(path))
  os.path.isfile = __isfile

  _islink = os.path.islink
  def __islink(path): return _islink(_realPath(path))
  os.path.islink = __islink

  _stat, _lstat = os.stat, os.lstat
  def __stat(path):  return  _stat(_realPath(path))
  def __lstat(path): return _lstat(_realPath(path))
  os.stat, os.lstat = __stat, __lstat

  _listdir = os.listdir
  def __listdir(path): return _listdir(_realPath(path))
  os.listdir = __listdir


  # Patch the open function
  class Open(object):
    def __init__(_, path, mode):
      _log.debug("calling patched 'open' function")
      _.path = path
      _.mode = mode
      _.fd = _open(_realPath(path), mode)

    def __enter__(_):
      _log.debug("Entering patched 'open' context manager")
      return _.fd

    def __exit__(_, *args):
      _log.debug(f"Exiting patched 'open' context manager {args}")
      try: _.fd.close()
      except: pass
      finally: del _.fd

    def read(_,  *args, **kwargs): return _.fd.read( *args, **kwargs)
    def write(_, *args, **kwargs): return _.fd.write(*args, **kwargs)
    def close(_, *args, **kwargs):
      try:                         return _.fd.close(*args, **kwargs)
      except: pass
      finally: del _.fd
  _open, open = open, Open

  _chdir = os.chdir
  def __chdir(path): return _chdir(_realPath(path))
  os.chdir = __chdir

  _mkdir = os.mkdir
  def __mkdir(path, mode = _RIGHTS): return _mkdir(_realPath(path), mode)
  os.mkdir = __mkdir

  _rmdir = os.rmdir
  def __rmdir(path): return _rmdir(_realPath(path))
  os.rmdir = __rmdir

  _makedirs = os.makedirs
  def __makedirs(path, mode = _RIGHTS, exist_ok = False):
    return _makedirs(_realPath(path), mode = mode, exist_ok = exist_ok)
  os.makedirs = __makedirs


  class TestRepoTestCase(unittest.TestCase):
    def testStuff(_):
      _.assertIsNot(_exists, os.path.exists)  # ensure monkey-patch worked
      _.assertTrue(os.path.exists("./_test-data"))  # directory check
      _.assertTrue(_exists("_test-data/d/a.b"))
      _.assertTrue(_exists("./_test-data/d/a.b"))
      _.assertTrue(os.path.exists("./_test-data/d/a.b"))  # file check
      _.assertTrue(os.path.exists("./_test-data/D/a.b"))
      _.assertFalse(os.path.exists("./_test-data/D/a.c"))
      _.assertEqual(0, os.stat("_test-data/D/A.B")[ST_SIZE])
      _.assertEqual(0, os.lstat("_test-DATA/d/A.B")[ST_SIZE])
      _.assertTrue(os.path.isdir("_test-data"))
      _.assertFalse(os.path.isdir("_test-data/d/A.B"))
      _.assertFalse(os.path.isdir("_test-data/d/A.c"))  # even if file doesn't exist
      _.assertTrue(os.path.isfile("_test-data/d/A.b"))
      _.assertFalse(os.path.isfile("_test-data/d/A.c"))  # even if file doesn't exist
      _.assertFalse(os.path.islink("_test-data/d"))
      _.assertFalse(os.path.islink("_test-data/d/a.b"))
      with open("_test-data/d/tmp", "w"): pass  # touch
      os.unlink("_test-data/d/tmp")
      _.assertFalse(os.path.exists("_test-data/d/tmp"))
      try: os.makedirs("_test-data/tmp/2", exist_ok = True)
      except Exception as E: _.fail(str(E))
      _.assertTrue(os.path.exists("_test-data/tmp/2"))
      try: os.makedirs("_test-data/tmp/2", exist_ok = False); _.fail("Should have thrown OSError")
      except OSError: pass
      _.assertTrue(os.path.exists("_test-data/tmp/2"))
      os.rmdir("_test-data/tmp/2")
      os.rmdir("_test-data/tmp")  # TODO fails because not empty? .tagsplorer.cfg
      _.assertFalse(os.path.exists("_test-data/tmp"))
      os.mkdir("_test-data/tmp2")
      _.assertTrue(os.path.exists("_test-data/tmp2"))
      os.rmdir("_test-data/tmp2")
      _.assertFalse(os.path.exists("./_test-data/tmp2"))


  def load_tests(loader, tests, ignore):
    ''' Queried by unittest. '''
    tests.addTests(doctest.DocTestSuite("tagsplorer.simfs"))
    return tests

else: SIMFS = False


if __name__ == '__main__':
  logging.basicConfig(level = logging.DEBUG if os.environ.get("DEBUG", "False").lower() == "true" else logging.INFO, stream = sys.stderr, format = "%(asctime)-23s %(levelname)-8s %(name)s:%(lineno)d | %(message)s")
  if sys.platform == 'win32': print("Testing on Windows makes no sense. This is a Windows file system simulator!"); exit(1)
  os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))  # main repo folder
  unittest.main()  # warnings = "ignore")
  for file in ("_X.x", "_x.X"): wrapExc(lambda: _unlink(file))



# coding=utf-8

''' tagsPlorer test suite  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

# HINT Set environment variable SKIP=true to avoid reverting test data prior to test run

import doctest, inspect, logging, os, subprocess, sys, unittest, traceback
from io import StringIO

sys.argv.append("--stdout")  # trigger only stdout output. option removed in tp to not interpret as exclusive <stdout> tag
from tagsplorer import constants, lib, simfs, tp, utils  # entire files

REPO = '_test-data'


def call(argstr, cwd = os.path.dirname(os.path.abspath(__file__))):
  ''' Run in a subprocess, no code coverage. '''
  return subprocess.Popen(argstr, cwd = cwd, shell = True, bufsize = 1000000, stdout = subprocess.PIPE, stderr = subprocess.STDOUT).communicate()[0].decode(sys.stdout.encoding)


def runP(argstr, repo = None):  # instead of script call via Popen, this allows coverage collection
  sys.argv = ["tp.py", "-r", repo if repo else REPO, "-i", repo if repo else REPO] + (["--simulate-winfs"] if simfs.SIMFS else []) + utils.safeSplit(argstr, " ")  # fake arguments
  def tmp():
    logFile.write("TEST: %s " % inspect.stack()[3].function + " ".join(sys.argv) + "\n")
    try: tp.Main().parse_and_run()
    except SystemExit as e: logFile.write(f"EXIT: {e.code}\n")
  res = wrapChannels(tmp)
  logFile.write(res)
  logFile.write("\n")
  return res


def wrapChannels(func):
  oldv, oldo, olde = sys.argv, sys.stdout, sys.stderr
  buf = StringIO()
  sys.stdout = sys.stderr = buf
  handler = logging.StreamHandler(buf)
#  debug, info, warn, error = map(lambda func: lambda *s: func(" ".join([str(e) for e in s])), [_log.debug, _log.info, _log.warning, _log.error])
  tp._log.addHandler(handler)
  lib._log.addHandler(handler)
  utils._log.addHandler(handler)
#  utils.debug, utils.info, utils.warn, utils.error =\
#  lib.debug, lib.info, lib.warn, lib.error =\
#  tp.debug, tp.info, tp.warn, tp.error = debug, info, warn, error
  try: func()
  except Exception as E: buf.write(str(E) + "\n"); traceback.print_exc(file = buf)
  finally:
    sys.argv, sys.stdout, sys.stderr = oldv, oldo, olde
    tp._log.removeHandler(handler)
    lib._log.removeHandler(handler)
    utils._log.removeHandler(handler)
  return buf.getvalue()


def setUpModule():
  ''' Run once before the entire test suite. '''
  global logFile
  logFile = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".testRun.log"), "w")


def tearDownModule():
  ''' Run once after the entire test suite. '''
  logFile.close()
  if not os.environ.get("SKIP", "False").lower() == "true":
    try: os.unlink(REPO + os.sep + constants.INDEX)
    except: pass
    if SVN: call(f'svn revert   "{REPO + os.sep + constants.CONFIG}"')
    else:   call(f'git checkout "{REPO + os.sep + constants.CONFIG}"')


class TestRepoTestCase(unittest.TestCase):
  ''' All tests are run through the command-line interface of tp.py. '''

  def setUp(_):
    ''' Run before each testCase. '''
    try: os.unlink(REPO + os.sep + constants.INDEX)
    except FileNotFoundError: pass  # if earlier tests finished without errors
    if SVN:  call(f'svn revert   "{REPO + os.sep + constants.CONFIG}"')
    else:    call(f'git checkout "{REPO + os.sep + constants.CONFIG}"')
    try: os.unlink(os.path.join(REPO, "tagging", "anyfile1"))
    except: pass
    _.assertIn("Updated configuration entry", runP("--set case_sensitive=True -v"))  # fixed value for reproducibility TODO allow all combinations on all platforms
    runP("-U")  # initial indexing, invisible

  def assertAllIn(_, what, where):
    ''' Assert all elements of what have an exact match in where. '''
    [_.assertIn(a, where) for a in what]

  def assertAllInAny(_, what, where):
    ''' Assert all elements of what are contained in at least one of where's entries. '''
    [_.assertTrue([a in entry for entry in where]) for a in what]  # TODO check if not same as above and used correctly

  def testSjoin(_):
    _.assertEqual("",    utils.sjoin())
    _.assertEqual("",    utils.sjoin(""))
    _.assertEqual("",    utils.sjoin("", ""))
    _.assertEqual("a b", utils.sjoin("a", "b"))

  def testFunctions(_):
    def x(a):
      if a == None: raise Exception("xany should not process this part!")
      return bool(a)
    _.assertTrue(tp.xany(x, [3, 3, 3, None]))
    _.assertFalse(tp.xany(x, []))
    _.assertTrue(utils.xall(x, [3, 3, 3]))
    _.assertFalse(utils.xall(x, [0, None]))
    _.assertTrue(utils.isFile("tests.py"))
    _.assertFalse(os.path.isdir("tests.py"))
    _.assertFalse(utils.isFile(os.getcwd()))
    _.assertTrue(os.path.isdir(os.getcwd()))
    x = [1, 2]
    i = id(x)
    _.assertTrue(i == id(utils.lappend(x, 3)))  # returns existing object
    _.assertTrue(i == id(utils.lappend(x, [4, 5])))  # ???
    _.assertEqual([1, 2, 3, 4, 5], x)
    _.assertEqual(0, utils.findIndexOrAppend([], 1))
    _.assertEqual(0, utils.findIndexOrAppend([1], 1))
    _.assertEqual(1, utils.findIndexOrAppend([1], 2))
    _.assertEqual([], utils.safeSplit(''))
    _.assertEqual(["1"], utils.safeSplit('1'))
    _.assertEqual(["1", "2"], utils.safeSplit('1,2'))
    _.assertEqual(["1", "2"], utils.safeSplit('1;2', ";"))
    d = utils.dd()
    d[1].append(1)
    _.assertEqual([1], d[1])

  def testGlobCheck(_):
    _.assertFalse(utils.isGlob(""))
    _.assertTrue(utils.isGlob("*"))
    _.assertTrue(utils.isGlob("???"))
    _.assertTrue(utils.isGlob("*a.c"))
    _.assertTrue(utils.isGlob("*a*.c"))
    _.assertTrue(utils.isGlob("*a*.c"))
    _.assertTrue(utils.isGlob("a??.c"))
    _.assertTrue(utils.isGlob("how.do?"))
    _.assertTrue(utils.isGlob("sbc.*"))
    _.assertFalse(utils.isGlob("sbc.a"))
    _.assertFalse(utils.isGlob("sbca"))

  def testSafesplit(_):
    _.assertEqual([], utils.safeSplit(""))
    _.assertEqual([], utils.safeSplit(","))
    _.assertEqual(["a"], utils.safeSplit("a"))
    _.assertEqual(["a"], utils.safeSplit("a,"))
    _.assertEqual(["a"], utils.safeSplit(",a"))
    _.assertEqual(["a"], utils.safeSplit(",a,"))
    _.assertEqual(["a", "b"], utils.safeSplit("a,b"))
    _.assertEqual(["a", "b"], utils.safeSplit("a;b", ";"))

  def testOnlySearchTerms(_):
    _.assertAllIn(["Found 1 files", "/b/b2/b2a/x.x"], runP(".x -v"))

  def testReduceCaseStorage(_):
    #_.assertIn("Added configuration entry: reduce_storage = False", runP("--set reduce_storage=False"))
    _.assertIn("tags: 49", runP("--stats"))  # only few upper-case entries exist, therefore no big difference if reduce_storage is used
    _.assertIn("Found 2 files in %d folders" % (2 if constants.ON_WINDOWS or simfs.SIMFS else 1), runP("Case -v"))  # contained in /cases/Case
    _.assertIn("Found 0 files", runP("CASE -v"))  # wrong case writing, can't find
    _.assertIn("Found %d files in %d folders" % ((4, 2) if constants.ON_WINDOWS or simfs.SIMFS else (2, 1)), runP("case -v -c"))  # ignore case: should find Case and case (no combination because different findFiles calls)
    _.assertIn("Added configuration entry", runP("--set reduce_storage=True -v"))
    _.assertAllIn(["Configuration entry: case_sensitive = True", "Configuration entry: reduce_storage = True"], runP("--config -v"))
    runP("-U")  # trigger update index after config change (but should automatically do so anyway)
    _.assertIn("tags: 49", runP("--stats"))  # now also small on Windows Windows
    _.assertIn("Found 2 files in 1 folders", runP("Case -v"))  # index contains original case only
    _.assertIn("Found 0 files in 0 folders", runP("case -v"))  # normalized version not in index anymore
#    _.assertIn("Found 2 files in ? folders", runP("case -c -v"))  # find anyway TODO should work but gets 0 in 0
    _.assertIn("Reset configuration parameters", runP("--reset"))  # TODO test if parameters reset to default

  def testFilenameCaseSetting(_):
    ''' This test confirms that case setting works (only executed on Linux). '''
    if utils.ON_WINDOWS or simfs.SIMFS: return  # TODO only skip the minimal part that is Linux-specific, but not all TODO or run always with simfs?
    _.assertIn("Found 0 files", runP("-s case --debug"))  # lower-case
    _.assertIn("Found 2 files", runP("-s Case --debug"))  # mixed-case
    _.assertIn("Found 0 files", runP("-s CASE --debug"))
    _.assertIn("Found 2 files", runP("-s case --debug --ignore-case"))  # should be same as next line above
    _.assertIn("Updated configuration entry", runP("--set case_sensitive=False -v"))
    _.assertIn("Wrote", runP("-U -v"))  # update after config change
    _.assertIn("Found 2 files", runP("-s Case -V"))
    _.assertIn("Found 2 files", runP("-s case -v"))  # TODO should be found no matter what
    # Now file-search
    _.assertIn("Found 1 files", runP("-s x.x -v"))  # doesn't find because case-normalized X.X doesn't exist
    _.assertIn("Found 1 files", runP("-s x.x -v"))  # TODO also test --force for removed file
    _.assertIn("Found 1 files", runP("-s X.x -v --ignore-case"))
    _.assertIn("Found 1 files", runP("-s x.x --ignore-case -v"))
    _.assertIn("Updated configuration entry", runP("--set case_sensitive=True"))  # revert
    _.assertIn("Wrote", runP("-U -v"))
    _.assertIn("Found 1 files", runP("-s x.x -v"))
    _.assertIn("Found 0 files", runP("-s X.x -v"))
    _.assertIn("Found 1 files", runP("-s X.x -v -c"))

  def testConfigs(_):
    ''' This test tests global configuration CRUD. '''
    _.assertAllIn(["Added configuration entry"], runP("--set __test=123 -v"))
    _.assertIn("Updated configuration entry", runP("--set __test=234 -v"))
    ret = runP("--get __test -v")
    _.assertIn("__test = 234", ret)
    _.assertIn("Configuration entry", ret)
    _.assertIn("Removed configuration entry", runP("--unset __test -v"))

  def testIllegalConfig(_):
    class MyIO(StringIO):
      def readlines(_):  return iter(_.read().split("\n"))
      def xreadlines(_): return _.readlines()
    def tmp():
      buf = MyIO("1494711739628\n[]\nfoo=bar\n")
      cp = lib.ConfigParser()
      cp.load(buf)
    res = wrapChannels(tmp)
    _.assertIn('Encountered illegal', res)
    def tmp2():
      buf = MyIO("1494711739628\n[]\nfoo\n")
      cp = lib.ConfigParser()
      cp.load(buf)
    res = wrapChannels(tmp2)
    _.assertIn('Key without value', res)

  def testGlobalIgnoreDir(_):
    _.assertAllIn(["Found 0 files"], runP("-s filea.exta -v"))  # was "No folder match" earlier, but searching files reduces the "includes" list to [] which returns all paths now
    _.assertNotIn("filea.exta", runP("-s filea.exta"))

  def testGlobalSkipDir(_):  # should skip /c/c2 which contains "filec.extb"
    _.assertIn("Found 0 files", runP("-s filec.extb -v"))  # should not been found due to skipd setting
    _.assertNotIn("filec.extb", runP("-s filec.extb"))

  def testLocalIgnoreDir(_):
    _.assertIn("Found 0 files", runP("-s 3.3 -v"))  # not filtering on folder tags
    _.assertIn("1.2", runP("-s 1.2 -v"))  #
    _.assertIn("2.1", runP("-s 2.1 -v"))
    _.assertIn("Found 0 files", runP("-s .3 -v"))  # due to local ignore marker file

  def testLocalSkipDir(_):
    _.assertIn("Found 0 files", runP("-s ignore_skip,marker-files,b,1.1 -v"))
    _.assertIn("Found 1 files", runP("-s ignore_skip,marker-files,b,2.2 -v"))
    _.assertNotIn("'.3'", runP("--stats -V"))
    _.assertIn("'.2'", runP("--stats -V"))

  def testLocalTag(_):
    _.assertIn("Found 1 folders", runP("-s b1,tag1 -v --dirs"))  #
    _.assertIn("Found 1 files in 1 folders", runP("-s b1,tag1 -v"))  # each pattern line matches exlusive sets, otherwise exclusion cannot work
    _.assertIn("Found 1 files in 1 folders", runP("-s b1 -s tag1 -v"))  # different interface, same result

  def testMappedInclude(_):
    _.assertAllIn(["Found 1 file", "/mapping/two/2.2"], runP("-s two,test -v"))  # finds folder two with a mapped direct tag test

  def testExtensionOnly(_):
    _.assertIn("Found 0 files in %d folders" % (27 if constants.ON_WINDOWS or simfs.SIMFS else 26), runP(".xyz -v"))  # /cases/case is added on Windows, and '' is not filtered on Windows

  def testMappedGlobExclude(_):
    pass

  def testOnlyDirsOption(_):
    _.assertAllIn(["Found 1 folder", "/folders/folder1"], runP("-s folder1 -v --dirs"))
    _.assertAllIn(["Found 3 folders", "/folders", "/folders/folder1", "/folders/folder2"], runP('-v -s folder? --dirs'))  # must be quoted on command-line, though
    _.assertAllIn(["Found 4 folders", "folder2", "folder1", "folders", "dot.folder"], runP('-s folder* -v --dirs'))

  def testExtensions(_):
    _.assertIn("Found 3 files in 2 folders", runP("-s .ext1 -v"))
    _.assertIn("more than one file extension", runP("-s .ext1 .ext2 -v"))
    _.assertIn("Found 1 files", runP("-s .ext1,extension -v"))

  def testFindFolder(_):
    def tmp():
      i = lib.Indexer(REPO)
      i.load(os.path.join(REPO, constants.INDEX), ignore_skew = True, recreate_index = False)
      print(i.findFolders(["folders", "folder2"], []))
    res = wrapChannels(tmp)
    _.assertAllIn(['/folders/folder2'], res)  # , 'Found potential matches in 1 folders'
    def tmp2():
      i = lib.Indexer(REPO)
      i.load(os.path.join(REPO, constants.INDEX), ignore_skew = True, recreate_index = False)
      print(utils.wrapExc(lambda: set(i.getPaths(i.tagdir2paths[i.tagdirs.index("a")], {})), lambda: set()))  # print output is captured
    _.assertAllIn(['/a/a2', '/a', '/a/a1', '/ignore_skip/marker-files/a'], wrapChannels(tmp2))

  def testStats(_):
    _.assertNotIn(" 0 occurrences", runP("--stats -v"))

  def testTokenization(_):
    _.assertAllIn(["Found 1 files",  "_test-data/dot.folder/one"], runP("folder -v"))  # find dot.folder via token
    _.assertAllIn(["Found 1 files",  "_test-data/dot.folder/one"], runP("dot -v"))
    _.assertAllIn(["Found 1 folder", "_test-data/ignore_skip"],    runP("skip -v --dirs"))

  def testGlobs(_):
    _.assertIn("Found 1 files in 1 folders", runP('dot.* -v'))

  def testInit(_):
    _.assertAllIn(["Create root configuration", "Wrote", "config bytes"], runP("-I -v --force", repo = os.path.join("_test-data", "tmp")))
    _.assertAllIn(["Index already exists", "--force"],                    runP("-I -v",         repo = os.path.join("_test-data", "tmp")))
    _.assertAllIn(["Create root configuration", "Wrote", "config bytes"], runP("-I -v --force", repo = os.path.join("_test-data", "tmp")))
    try: os.unlink(os.path.join("_test-data", "tmp", ".tagsplorer.cfg")); os.rmdir(os.path.join("_test-data", "tmp"))
    except Exception as E: _.fail(str(E))

  def testAddRemove(_):
    ''' Add a tag, check and remove. '''
    _.assertIn("Found 0 files", runP("-s missing -v"))
    _.assertIn("No file matches", runP("--tag missing,-exclusive _test-data/tagging/anyfile1 -v"))
    _.assertNotIn("Wrote", runP("--tag missing,-exclusive _test-data/tagging/anyfile1 -v"))
    _.assertAllIn(["folders", "folder2"], runP("--tags _test-data/folders/folder2"))
    _.assertNotIn("Recreate", runP("-s missing -v"))
    _.assertIn("Found 0 files", runP("-s missing -v"))  # should still not find anything TODO ??
    # test adding non-existing file, then search
    _.assertAllIn(["Tag <missing> in '/tagging' for +anyfile1 -", "add anyway"], runP("--tag missing,-exclusive _test-data/tagging/anyfile1 -v --force"))
    _.assertIn("Found 0 files in 1 folders", runP("-s missing -v"))
    _.assertIn("remove anyway", runP("--untag missing,-exclusive _test-data/tagging/anyfile1 -v --force"))
    _.assertIn("Found 0 files", runP("-s missing -v"))  # because neither file nor tagging exists
    # test adding on existing file, then search
    with open(os.path.join(REPO, "tagging", "anyfile1"), "w") as fd: fd.close()  # touch to create file
    _.assertIn("Tag <missing> in '/tagging' for +anyfile1 -", runP("--tag missing,-exclusive _test-data/tagging/anyfile1 -v"))
    _.assertIn("Found 1 files in 1 folders", runP("-s missing -v"))  # is now found, since existing
    try: os.unlink(os.path.join(REPO, "tagging", "anyfile1"))  # remove again
    except: pass
    _.assertIn("skip", runP("--untag missing,-exclusive _test-data/tagging/anyfile1 -v"))
    _.assertIn("anyway", runP("--untag missing,-exclusive _test-data/tagging/anyfile1 -V --force"))
    _.assertIn("Found 0 files", runP("-s missing -v"))

  def testNoOption(_):
    _.assertIn("No option", runP(""))

  def testNegativeSearch(_):
    _.assertAllInAny(["Found 4 folders for +<a> -<>", "/a", "/a/a1", "/a/a2"], runP("-s a -v --dirs").split("\n"))  # only display dirs
    _.assertAllInAny(["Found 3 folders for +<a> -<a1>", "/a", "/a/a2"], runP("-s a -x a1 -v --dirs").split("\n"))  # with exclude only dirs
    _.assertIn("Found 8 files in 4 folders", runP("-s a -v"))  # only include with files
    _.assertAllIn(["Found 4 files in 3 folders", "file3.ext1", "file3.ext2", "file3.ext3"], runP("-s a -x a1 -v"))  # with exclude with files

  @unittest.skip("doesn't run on CI because No module named 'tagsplorer'")
  def testTestLib(_):
    _.assertAllIn(["Test passed", "Test passed"], call(PYTHON + " lib.py -v", cwd = os.path.dirname(os.path.abspath(__file__)) + os.sep + "tagsplorer"))
    _.assertAllIn(["Test passed", "Test passed"], call(PYTHON + " tp.py --test -v", cwd = os.path.dirname(os.path.abspath(__file__)) + os.sep + "tagsplorer"))
    _.assertAllIn(["Test passed", "Test passed"], call(PYTHON + " utils.py -v", cwd = os.path.dirname(os.path.abspath(__file__)) + os.sep + "tagsplorer"))

  def testExtensionAndTag(_):
    _.assertAllIn(["Found 2 files in 1 folders", "/b/b1/file3.ext1"], runP("b .ext1 -v"))
    _.assertAllIn(["Found 2 files in 1 folders", "/b/b1/file3.ext1"], runP(".ext1 b -v"))  # should be same (regression test)
    _.assertAllIn(["Found 0 files"], runP("a .3 -v"))  # doesn't exist
    _.assertAllIn(["Found 0 files"], runP("a .2 -v"))  # while a lists all files in folder, filtering with .2 doesn't match any a
    _.assertAllIn(["Found 1 folders", "/b/b1"], runP(".ext1 b -v --dirs"))  # was returning all folders previously

  def testNegativeExtension(_):
    _.assertIn("more than one file extension", runP("a,-.ext1,-.ext2 -v"))
    _.assertIn("Found 6 files in 4 folders", runP("-s a -x .ext1 -v"))
    _.assertNotIn(".ext1", runP("-s a -x .ext1"))
    _.assertIn("Found 4 files in 1 folders", runP("a1 -x .ext1 -v"))
    _.assertIn("Found 3 files in 3 folders", runP("-s a -x .ext2 -v"))
    _.assertIn("Found 2 files in %d folders" % (27 if constants.ON_WINDOWS or simfs.SIMFS else 26), runP("b1 -x .ext2 -v"))

  def testUnwalk(_):
    def unwalk(_, idx = 0, path = ""):
      ''' Walk entire tree from index (slow but proof of correctness). '''
      tag = _.tagdirs[idx]  # name of head element
      children = (f[0] for f in filter(lambda a: a[1] == idx and a[0] != idx, ((e, v) for e, v in enumerate(_.tagdir2parent))))  # using generator expression
      print(path + tag + constants.SLASH)
      for child in children: unwalk(_, idx = child, path = path + tag + constants.SLASH)

    def tmp():
      i = lib.Indexer(REPO)
      i.load(os.path.join(REPO, constants.INDEX), True, False)
      unwalk(i)
    result = wrapChannels(tmp).replace("\r", "")
    logFile.write(result + "\n")
    _.assertEqual(31, len(result.split("\n")))  # TODO check if correct


def _compressionTest():
  ''' This is not a unit test, rather a benchmark test code. '''
  import timeit
  i = lib.Indexer("../..")
  for j in range(10):
    print(f"Compression level {j}")
    i.compressed = j
    i.store("../_test-data/" + constants.INDEX)
    s = os.stat(constants.INDEX)[6]  # get compressed size
    print("Level %d: %f %d" % (j, timeit.Timer(lambda: i.load(constants.INDEX)).timeit(number = 20), s))


def load_tests(loader, tests, ignore):
  ''' Added up by unittest. '''
  tests.addTests(doctest.DocTestSuite(lib))
  tests.addTests(doctest.DocTestSuite(simfs))
  tests.addTests(doctest.DocTestSuite(tp))
  tests.addTests(doctest.DocTestSuite(utils))
  return tests


if __name__ == '__main__':
  os.chdir(os.path.dirname(os.path.abspath(__file__)))  # change to own folder
  PYTHON = '"' + os.path.realpath(sys.executable) + '"' if constants.ON_WINDOWS else os.path.realpath(sys.executable)
  logFile = None
  SVN = tp.findRootFolder('.svn') is not None
  print(f'Using VCS \'{"Subversion" if SVN else "Git"}\' to revert test data')
  unittest.main(exit = False)  # warnings = "ignore")



''' tagsPlorer setup script  (C) 2017-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

import os, shutil, subprocess, sys, time
from setuptools import setup

from tagsplorer.constants import MAJOR_VERSION

if os.path.exists(".git"):
  so, se = subprocess.Popen("git describe --always", shell = True, stdout = subprocess.PIPE).communicate()
  micro = "-" + so.strip().decode(sys.stdout.encoding).strip()
else: micro = ""
lt = time.localtime()
versionString = "%d.%d.%d" % (MAJOR_VERSION, lt.tm_year * 100 + lt.tm_mon, lt.tm_mday * 10000 + lt.tm_hour * 100 + lt.tm_min) if any(_ in sys.argv for _ in ('clean', 'build')) else open(os.path.join(os.path.dirname(os.path.join(__file__)), "tagsplorer", "VERSION"), "r", encoding = "utf-8").read()
if 'clean' not in sys.argv:
  with open(os.path.join(os.path.dirname(os.path.join(__file__)), "tagsplorer", "VERSION"), "w", encoding = "utf-8") as fd: fd.write(versionString)


setup(
  name = 'tagsPlorer',
  version = versionString,
  description = "tagsPlorer V" + versionString + micro,
  description_content_type = "text/markdown",
  long_description = "",
  long_description_content_type = "text/markdown",
  classifiers = [c.strip() for c in """
        Development Status :: 5 - Production/Stable
        Intended Audience :: Other Audience
        Intended Audience :: System Administrators
        License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)
        Operating System :: OS Independent
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.6
        Programming Language :: Python :: 3.7
        Programming Language :: Python :: 3.8
        Programming Language :: Python :: 3.9
        """.split('\n') if c.strip()],  # https://pypi.python.org/pypi?:action=list_classifiers
  keywords = '',
  author = 'Arne Bachmann',
  author_email = 'ArneBachmann@users.noreply.github.com',
  maintainer = 'Arne Bachmann',
  maintainer_email = 'ArneBachmann@users.noreply.github.com',
  url = 'http://github.com/ArneBachmann/corrupdetect',
  license = 'MPL-2.0',
  packages = ["tagsplorer"],
  #package_dir = {"tagsplorer": ""},
  package_data = {"tagsplorer": ["VERSION"]},
  zip_safe = False,
  entry_points = {
    'console_scripts': [
      'tp=tagsplorer.tp:main'
    ]
  },
)

if "clean" in sys.argv:
  try: shutil.rmtree("tagsplorer.egg-info")  # if keeping this folder, built files will remain no matter what exclude options are configured above (MANIFEST.in or exclude_...)
  except: pass  # noqa: E722
  try: shutil.rmtree("build")
  except: pass  # noqa: E722
  try: shutil.rmtree("dist")
  except: pass  # noqa: E722



''' tagsPlorer utilities  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

import collections, fnmatch, logging, os, re, sys, time
from functools import reduce

from tagsplorer.constants import COMB, ON_WINDOWS, REGEX_SANITIZE, SKIP, SLASH


_log = logging.getLogger(__name__)
def log(func): return (lambda *s: func(sjoin([_() if callable(_) else _ for _ in s]), **({"stacklevel": 2} if sys.version_info >= (3, 8) else {})))
debug, info, warn, error = log(_log.debug), log(_log.info), log(_log.warning), log(_log.error)


pathNorm = (lambda s: s.replace(os.sep, SLASH)) if ON_WINDOWS else lambda _: _  # HINT do not convert into a function to allow ternary expression here
casefilter = lambda lizt, pat: [name for name in lizt if fnmatch.fnmatchcase(name, pat)]  # case-sensitive filter


class Normalizer(object):
  ''' Class that provides all case-normalization functions. '''

  def setupCasematching(_, case_sensitive, suppress = False):
    ''' Setup normalization.
    >>> n = Normalizer(); n.setupCasematching(case_sensitive = True)
    >>> print((n.filenorm("Abc"), n.globmatch("abc", "?Bc"), n.globfilter(["ab1", "Ab2"], "a??")))
    ('Abc', False, ['ab1'])
    >>> print(n.globfilter(["ab1", "Ab2"], "a*"))
    ['ab1']
    >>> n.setupCasematching(False)
    >>> print((n.filenorm("Abc"), n.globmatch("abc", "?Bc"), n.globfilter(["ab1", "Ab2"], "a??")))
    ('abc', True, ['ab1', 'Ab2'])
    >>> print(n.globfilter(["ab", "Ab"], "a?"))
    ['ab', 'Ab']
    >>> print(n.globfilter(["dot.folder"], "DOT*"))  # regression test
    ['dot.folder']
    '''
    if not suppress: debug("Case-sensitive matching: %s" % ("On" if case_sensitive else "Off"))
    _.filenorm   = (lambda _: _)       if case_sensitive else str.lower
    _.globmatch  = fnmatch.fnmatchcase if case_sensitive else lambda f, g: fnmatch.fnmatch(f.lower(), g.lower())
    _.globfilter = casefilter if case_sensitive else \
                   (lambda lizt, pat: [name for name in lizt if fnmatch.fnmatch(name.lower(), pat.lower())])  # HINT lower maybe not necessary
normalizer = Normalizer()  # keep a static module-reference


def dd(): return collections.defaultdict(list)


def safeSplit(s, d = COMB):
  ''' Split a string removing empty sub-strings.
  >>> safeSplit("a,b,,d")
  ['a', 'b', 'd']
  >>> safeSplit(",x,")
  ['x']
  >>> safeSplit(',')
  []
  '''
  return [_.strip() for _ in s.split(d) if _ != '']  # remove empty strings that appear e.g. for "  ".split(" ") or "".split(" ")


def safeRSplit(s, d = SLASH):
  ''' A right-split that avoids an error if the delimiter is not contained.
  >>> safeRSplit('a')
  'a'
  >>> safeRSplit('/a/')
  ''
  >>> safeRSplit('/a')
  'a'
  '''
  return s[s.rindex(d) + 1:] if d in s else s


def sjoin(*string, sep = " "):
  ''' Join strings.
  >>> sjoin("a", "b")
  'a b'
  >>> sjoin([1, 2, 3])
  '1 2 3'
  >>> sjoin(["a", None, 2, ""])
  'a 2'
  '''
  if not string: return ""
  string = string[0] if isinstance(string[0], (list, set)) else string
  return sep.join([str(elem) for elem in string if elem])


# short-circuit operations: intentionally doesn't iterate over string characters. Converts string and other data types to a one-element list instead
def xany(pred, *lizt): return reduce(lambda a, b: a or  pred(b), lizt[0] if isinstance(lizt[0], (list, set, dict)) else lizt, False)
def xall(pred, *lizt): return reduce(lambda a, b: a and pred(b), lizt[0] if    hasattr(lizt[0], '__iter__')        else lizt, True)


def wrapExc(func, otherwise = None):
  ''' Wrap a no-args function; catch any exception and compute return value lazily.
  >>> print(wrapExc(lambda: 1))
  1
  >>> None is wrapExc(lambda: 1 / 0)  # return default fallback value (None)
  True
  >>> print(wrapExc(lambda: 1 / 0, lambda: 2))  # return defined default by function call
  2
  '''
  try: return func()
  except Exception: return otherwise() if callable(otherwise) else otherwise


def appendnew(lizt, value):
  ''' Only add if not yet contained.
  >>> print(appendnew([1, 2], 3))
  [1, 2, 3]
  >>> appendnew([1, 2], 1)
  [1, 2]
  '''
  if value not in lizt: lizt.append(value)
  return lizt


def lindex(lizt, value, otherwise = lambda _lizt, _value: None):
  ''' List index function with lazy append computation.
      Returns index if found, otherwise appends and returns new index
      Default fallback logic if no computation function is given: do nothing and return None.
  >>> print(lindex([1, 2], 2))  # returns index of value 2 in search array
  1
  >>> print(lindex([], 1))  # default not-found computation:
  None
  >>> print(lindex([1], 2, lambda _lizt, _value: _lizt.index(_value - 1)))  # computes index of previous value
  0
  '''
  return wrapExc(lambda: lizt.index(value), lambda: otherwise(lizt, value))  # ValueError


def lappend(listOrSet, elem):
  ''' Append one or more element(s) to a set or list, returning the updated set or list (in-place).
  >>> a = [1, 2, 3]; lappend(a, 4)  # check return list
  [1, 2, 3, 4]
  >>> a == [1, 2, 3, 4]  # was updated in-place
  True
  >>> lappend(a, [1, 2])  # append several values
  [1, 2, 3, 4, 1, 2]
  >>> list(sorted(lappend(set([1]), {2, 0})))  # append using a set
  [0, 1, 2]
  '''
  assert isinstance(listOrSet, (list, set))
  ((listOrSet.extend if isinstance(listOrSet, list) else listOrSet.update) if isinstance(elem, (list, set)) else (listOrSet.append if isinstance(listOrSet, list) else listOrSet.add))(elem)
  return listOrSet


def findIndexOrCreateNext(lizt, elem):  # TODO unused
  ''' Returns index of element or next available index. This differs from lindex by providing the "otherwise next" part.
  >>> findIndexOrCreateNext([1, 2, 5], 4)
  3
  >>> findIndexOrCreateNext([1, 2, 5], 2)
  1
  '''
  assert isinstance(lizt, list)
  assert not isinstance(elem, (list, set))
  return lindex(lizt, elem, otherwise = lambda l, v: len(l))


def findIndexOrAppend(lizt, elem):
  ''' Find index of element if existing, else add and return its index.
  This differs from findIndexOrCreateNext by actually appending the elemnt to the list.
  >>> l = [1, 2, 3]; findIndexOrAppend(l, 2)
  1
  >>> l = [1, 2, 3]; findIndexOrAppend(l, 4)
  3
  >>> print(l)
  [1, 2, 3, 4]
  '''
  return lindex(lizt, elem, otherwise = lambda l, v: l.append(v) or len(l) - 1)


def isDir(f):  return wrapExc(lambda: os.path.isdir(f) and not os.path.ismount(f), False)  # HINT silently catches encoding errors


def isFile(f): return wrapExc(lambda: os.path.isfile(f), False)  # handle "no file" errors


def isGlob(f):
  ''' Is the specified pattern a potential glob?
      HINT TODO doesn't allow yet square bracket patterns like "[123] or "[a-z]""
  >>> isGlob("a*b.jp?")
  True
  >>> isGlob("sdf.txt")
  False
  >>> isGlob("folder?")
  True
  '''
  return '*' in f or '?' in f


def isUnderRoot(root, folder):
  r''' Check if a filepath is under the discovered repository root.
  >>> isUnderRoot("C:\\", "D:\\file.txt")
  False
  >>> isUnderRoot("C:\\Windows", "C:\\Windows\\System")
  True
  >>> isUnderRoot("C:\\Windows", "C:\\Windows")  # same path
  True
  '''
  return os.path.commonprefix([root, folder]).startswith(root)



def getTsMs(): return int(time.time() * 1000.)  # timestamp rounded to milliseconds for metadata comparison


def dictGet(dikt, key, default, set = False):
  ''' dict.get() that returns default when missing.
      dikt: a dictionary
      key:  key to find value of
      default: value to use if key missing
  >>> dictGet({}, 'a', dictGet({'a': 1}, 'b', 2))
  2
  '''
  try: return dikt[key]
  except KeyError:
    if set: dikt[key] = default
    return default


def dictGetSet(dikt, key, default):
  ''' dict.get() that computes and adds missing values.
      Returns existing value otherwise new value.
  >>> a = {}; b = a.get(0, 0); print((a, b))  # normal dict get
  ({}, 0)
  >>> a = {}; b = dictGetSet(a, 0, 0); print((a, b))  # improved dict get
  ({0: 0}, 0)
  '''
  return dictGet(dikt, key, default, set = True)


def splitByPredicate(lizt, pred):
  ''' Split lists by a (boolean) predicate.
      returns ([if-true], [if-false])
  >>> print(splitByPredicate([1,2,3,4,5], lambda e: e % 2 == 0))
  ([2, 4], [1, 3, 5])
  '''
  MATCH, NO_MATCH = 0, 1
  return reduce(lambda acc, nxt_: (
      lappend(acc[MATCH], nxt_), acc[NO_MATCH])
    if pred(nxt_) else
      (acc[MATCH], lappend(acc[NO_MATCH], nxt_)),
    lizt, ([], []))


def caseCompareKey(c):
  ''' Python 3 key function for sorting, only used in debug output. '''
  if c == '': return 0
  if len(c) > 1: return ord(c[0].lower()) << 8 | caseCompareKey(c[1:])
  return ord(c.lower())


def escapeRegex(name):
  r''' Escape the string for regex matching.
  >>> escapeRegex("a(b)")
  'a\\(b\\)'
  '''
  return REGEX_SANITIZE.sub(r"\\\1", name)


def pathHasGlobalSkip(path, skips):
  ''' True if any path matches the global skip paths (regex for full, beginning, middle, end).
      Uses dynamic generation of regular expressions TODO path sanitization before re for security
  >>> pathHasGlobalSkip('/a', ['a'])
  True
  >>> pathHasGlobalSkip('/b', ['a'])
  False
  >>> pathHasGlobalSkip('/a/b/c', ['b'])
  True
  '''
  assert isinstance(path, str)
  assert isinstance(skips, list)
  return xany(lambda skip: wrapExc(lambda: re.search(r"((^%s$)|(^%s/)|(/%s/)|(/%s$))" % ((escapeRegex(skip),) * 4), path).groups()[0].replace("/", "") == skip, False), skips)


def pathHasGlobalIgnore(path, ignores):
  ''' Check if last path constituent matches any global ignore pattern.
  >>> normalizer.setupCasematching(True)
  >>> pathHasGlobalIgnore('/a/b', ['b'])
  True
  >>> pathHasGlobalIgnore('/a/b', ['B'])
  False
  >>> normalizer.setupCasematching(False)
  >>> pathHasGlobalIgnore('/a/b', ['c'])
  False
  >>> pathHasGlobalIgnore('/a/b', ['C'])
  False
  >>> pathHasGlobalIgnore('', [''])
  True
  '''
  assert isinstance(path, str)
  assert isinstance(ignores, list)
  if not path: return '' in ignores  # check root ignore
  basename = path[path.rindex(SLASH) + 1:] if SLASH in path else path
  return xany(lambda ignore: normalizer.globmatch(basename, ignore), ignores)


def anyParentIsSkipped(path, paths):
  ''' True if any prefix path of path is marked as "skip", so that all chlildren can be skipped, too.
  >>> anyParentIsSkipped('/a/b', {'/a': {'skip': ''}})
  True
  >>> anyParentIsSkipped('/a/b', {'/a/b': {'ignore': ''}})
  False
  '''
  assert isinstance(path, str)
  assert isinstance(paths, dict)
  return xany(lambda i: SKIP in dictGetSet(paths, SLASH.join(path.split(SLASH)[:i + 1]), {}), list(range(path.count(SLASH))))  # test all parent paths


def splitTags(lizt):
  ''' Parse comma-separated arguments like tags.
  >>> print(splitTags([]))
  []
  >>> print(splitTags(["a,b"]))
  ['a', 'b']
  >>> print(splitTags(["ab"]))
  ['ab']
  >>> print(splitTags(["ab,"]))
  ['ab']
  '''
  assert isinstance(lizt, list)
  return [x for x in (y.strip() for y in reduce(lambda a, b: a + ([b] if COMB not in b else safeSplit(b)), lizt, [])) if x != ""]


def removeTagPrefixes(poss, negs):
  ''' Remove leading "inclusive" or "exclusive" +/- markers from tag lists.
  >>> print(removeTagPrefixes(["a"], []))
  (['a'], [])
  >>> print(removeTagPrefixes(["+a", "b"], ["-c", "d"]))
  (['a', 'b'], ['c', 'd'])
  '''
  return [p.lstrip('+') for p in poss], [n.lstrip('-') for n in negs]


if __name__ == '__main__': import doctest; doctest.testmod()



''' tagsPlorer web portal  (C) 2016-2021  Arne Bachmann  https://github.com/ArneBachmann/tagsplorer '''

# TODO: second index listing is indented
# IDEA: render site, show hourglass and return search result asynchronously (requires spawning separate process, difficult to keep track of?)
# TODO: on timeout inside index still return anything - needs refactoring in findFolders/findFiles
# TODO: add HTTPS https://github.com/goya191/SimpleAuthServerSSL.py/blob/master/SimpleHttpsAuthServer.py

import base64, hashlib, logging, os, sys, socket, time, urllib, urllib2

logging.basicConfig(
  level =  logging.DEBUG if '-V' in sys.argv or '--debug'   in sys.argv or os.environ.get("DEBUG",   "False").lower() == "true" else
          (logging.INFO  if '-v' in sys.argv or '--verbose' in sys.argv or os.environ.get("VERBOSE", "False").lower() == "true" else
           logging.WARNING),  # must be set here to already limit writing to info and debug (e.g. "Started at...")
  stream = sys.stdout if '--stdout' in sys.argv else sys.stderr,  # log to stderr, write results to stdout
  format =  '%(asctime)-5s.%(msecs)03d  %(levelname)-7s  %(module)s:%(funcName)s:%(lineno)d | %(message)s',
  datefmt = '%M:%S')

from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler

from tagsplorer import constants
from tagsplorer.lib import Indexer
from tagsplorer.utils import safeSplit, sjoin, splitByPredicate

_log = logging.getLogger(__name__)
def log(func): return (lambda *s: func(sjoin([_() if callable(_) else _ for _ in s]), **({"stacklevel": 2} if sys.version_info >= (3, 8) else {})))
debug, info, warn, error = log(_log.debug), log(_log.info), log(_log.warning), log(_log.error)

# Custom imports of tagsPlorer
from tp import commaArgsIntoList, DOT, findRootFolder, isglob, removeTagPrefixes, withoutFilesAndGlobs, xany


# Globals
LOGIN = 'ly'
PASSWORD = 'search'
SERVER = ('', 8000)  # localhost with given port
# SEARCHER = ('', 8001)  # asynchronous search server
HTTP = "http"  # folder with resources to serve (and nothing else)
CACHE = os.path.join(HTTP, "cache")
RESOURCES = set([file for file in os.listdir(HTTP)])  # resources allowed to get (whitelist to prevent attacks)
MAX_WAIT = 3.  # maximum time in seconds per query (for all indexes in total)
MAX_SHOW = 250  # maximum number of items to show
CREDENTIALS = base64.b64encode("%s:%s" % (LOGIN, PASSWORD))
MASK =\
  '''
  <form action="/search" method="get" name="search">
  <p>Search terms:
  <input type="text" name="q" id="query" value="%s" />&nbsp;
  Case-sensitive:&nbsp;
  <input type="checkbox" name="case" value="1" %s />
  <input type="submit" />&nbsp;
  <!-- span class="query-hint">Hint: Separate search terms by comma, exclude terms with minus, use dot to find file extensions, use ? and * for file globbing (slow).</span -->
  </p>
  </form>
  '''
mime = {
  "html": "text/html",
  "css": "text/css",
  "js": "application/javascript",
  "pdf": "application/pdf",
  "png": "application/png"
}

# Functions
def sanitize(s): return "".join([_ for _ in s if _ in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.*,"])  # s.replace("&", "&amp;")

def getResource(suffix): return 'custom' + suffix if 'custom' + suffix in RESOURCES else 'default' + suffix

def header():
  with open(os.path.join(HTTP, getResource("_header.xhtml")), 'rb') as fd: return fd.read().replace("${JS}", getResource(".js")).replace("${CSS}", getResource(".css"))

def title(name): return '<h1>' + name + '</h1>'

def footer():
  with open(os.path.join(HTTP, getResource("_footer.xhtml")), 'rb') as fd: return fd.read()


# Main class
class Handler(SimpleHTTPRequestHandler):

  def __init__(_, request, address, server):
    debug("Request from " + str(address))
    try: SimpleHTTPRequestHandler.__init__(_, request, address, server)
    except socket.error as E: error("Connection error, probably aborted by client: " + str(E))

  def log_message(_, format, *args):
    info(" ".join(args))

  def startTime(_):
    _.time = time.time()

  def endTime(_):
    info("Finished after %.3fms" % (1000. * (time.time() - _.time)))

  def end_headers(_, filename = ""):
    _.send_header('Content-type', mime.get(filename[filename.index(".") + 1:] if "." in filename[1:1] else "html", "html"))
    BaseHTTPRequestHandler.end_headers(_)

  def do_AUTHHEAD(_):
    _.send_response(401)
    _.send_header('WWW-Authenticate', 'Basic realm="LY datalake file search"')
    _.end_headers()

  def do_POST(_):
    _.send_response(405); _.end_headers()
    _.wfile.write(header() + title("POST queries not supported.") + MASK % ('', '') + footer())

  def do_GET(_):
    _.startTime()
    if _.headers.getheader('Authorization') == None:  # no authentication posted yet: ask user
      _.do_AUTHHEAD()
      _.wfile.write('<h1>No login credentials received in authorization header.</h1>')
      _.endTime()
      return
    elif _.headers.getheader('Authorization') != 'Basic ' + CREDENTIALS:  # wrong username/password combination entered: ask again
      try: info("Login attempt from %r" % base64.b64decode(_.headers.getheader('Authorization').split(" ")[1]))
      except: error("Couldn't decode authorization string")
      _.do_AUTHHEAD()
      _.wfile.write('<h1>Wrong credentials received in authorization header.</h1>')
      _.endTime()
      return
    # From here on, we're authorized
    if _.path in ["", "/"]:  # start page
      _.send_response(200); _.end_headers()
      _.wfile.write(header() + title("Perform a search") + MASK % ('', '') + footer())

    elif _.path[0] == '/' and _.path[1:] in RESOURCES:  # get resource
      if not os.path.exists(os.path.join(HTTP, _.path[1:])):
        _.send_response(404); _.end_headers()
        _.wfile.write(header() + title("Perform a search") + MASK % ('', '') + footer())
        _.endTime()
        return
      _.send_response(200); _.end_headers(_.path)
      with open(os.path.join(HTTP, _.path[1:])) as fd: _.wfile.write(fd.read())

    elif _.path.startswith("/search?q="):  # perform query
      global hashcache
      _.send_response(200); _.end_headers()
      search = _.path[_.path.index("=") + 1:]
      case = "&" in search and search.split("&")[1] == 'case=1'  # other parameters follow
      if case: search = search.split("&")[0]  # isolate search term itself
      search = urllib.unquote(search)
      if len(search) == 0: _.wfile.write(header() + title("@%s &rarr; No search term(s) given") % time.strftime("%H:%M:%S") + MASK % ('', 'checked="checked"' if case else '') + footer()); _.endTime(); return

      # Perform search
      search = safeSplit(sanitize(search).replace(" ", ","))
      debug("Sanititzed search terms (unsorted): %s" % search)
      poss, negs = splitByPredicate(commaArgsIntoList(search), lambda e: e[0] != '-')
      poss, negs = removeTagPrefixes(poss, negs)
      cacheString = ",".join(sorted(poss) + ["_"] + sorted(negs) + ["+" if case else "-"])
      hasher = hashlib.md5(); hasher.update(cacheString)  # create unique order of search terms
      cachehash = hasher.hexdigest()[:8]
      debug("Cache key and hash for search terms (sorted): '%s' (%s)" % (cacheString, cachehash))
      cachefile = os.path.join(CACHE, cachehash)
      if cachehash in hashcache and os.path.exists(cachefile):
        debug("Sending cached result")
        with open(cachefile, 'rb') as fd: _.wfile.write(fd.read()); debug("Finished sending cached result for '%s' (%s)" % (cacheString, cachehash))
        _.endTime()
        return
      with open(cachefile, 'wb') as fd:  # create new cache entry while searching TODO may be locked by other process: wait and read contents
        debug("Starting search for " + cacheString)
        hashcache.add(cachehash)  # memorize new entry as "known"
#        poss, negs = removeTagPrefixes(poss, negs)
        tmp = header() + "<p>Search criteria:\n<ul><li><span class='cursive'>Must match:</span> %s</li><li><span style='font-style:italic'>Must not match:</span> %s</li></ul></p>" % (", ".join(['"' + p + '"' for p in poss]), ", ".join(['"' + n + '"' for n in negs])) + MASK % (",".join(search), 'checked="checked"' if case else '')
        tmp += title("Search results")
        _.wfile.write(tmp); fd.write(tmp); _.wfile.flush()  # TODO check if flush really speeds up browser rendering?
        for repo in idx:  # for every repository
          init = time.time()
          shown = 0
          debug("Finding folders on '%s'..." % repo.root)
          repo.cfg.case_sensitive = case  # update search option flag (in-memory only, not persisting)
          folders = repo.findFolders(withoutFilesAndGlobs(poss), withoutFilesAndGlobs(negs))
          if len(folders) == 0 and xany(lambda x: isglob(x) or DOT in x, poss + negs): folders = repo.findFolders([], [], True)  # return all folders names unfiltered (except ignore/skip without marker files)
          skipped = []
          foundcount = 0
          debug("Filtering folders... ")
          for folder, (files, skip) in ((path, repo.findFiles(path, poss, negs)) for path in folders):
            if skip: skipped.append(folder); debug("Storing skip information for %r" % folder); continue  # memorize to skip all folders with this path prefix (from skip marker)
            if any([folder.startswith(skp) if skp != '' else (folder == '') for skp in skipped]): debug("Skipping folder", folder); continue  # is in skipped folder tree
            if len(files) > 0: foundcount += 1  # tmp = "<p class='mono fat'>" + repo.root + folder + "</p>\n<ul>"; _.wfile.write(tmp); fd.write(tmp)  # found something in this path
            for file in files: tmp = "<p class ='mono fat'>%s<a href='file://%s'>%s</a></p>" % (folder + os.sep, urllib2.quote(repo.root + folder[1 if repo.root.endswith("/") else 0:] + os.sep + file), file); _.wfile.write(tmp); fd.write(tmp)
            shown += len(files)
            if shown > MAX_SHOW: tmp = "<p class='fat emph'>@%s &rarr; Too many matches for %s, aborting search of this index for now.</p>" % (time.strftime("%H:%M:%S"), repo.root); _.wfile.write(tmp); fd.write(tmp); break
            if len(files) > 0: tmp = "</ul>"; _.wfile.write(tmp); fd.write(tmp); _.wfile.flush()
            if (time.time() - init) > MAX_WAIT / len(idx): tmp = "<p class='fat emph'>@%s &rarr; Database timed out for <span class='mono'>%s</span></p>" % (time.strftime("%H:%M:%S"), repo.root); _.wfile.write(tmp); fd.write(tmp); break  # folder loop
          if foundcount == 0: tmp = "<p class='fat emph'>@%s &rarr; Nothing found for <span class='mono'>%s</a></p>" % (time.strftime("%H:%M:%S"), repo.root); _.wfile.write(tmp); fd.write(tmp)
        tmp = footer(); _.wfile.write(tmp); fd.write(tmp)
      debug("Finished search for " + ",".join(search))
    else:  # unknown REST path
      warn("SPAM REQUEST: " + repr(_.path))
      _.send_response(404); _.end_headers()
      _.wfile.write(header() + title("@%s &rarr; No access to this URL, please use provided search box") % time.strftime("%H:%M:%S") + MASK % ('', '') + footer())
    _.endTime()


# Main entry point
if __name__ == '__main__':
  if len(sys.argv) < 2 or "--help" in sys.argv:
    print("serve.py repopath1,configpath1 repopath2,configpath2 ...")
    sys.exit(0)
  if len(sys.argv) > 1:
    REPOS, METAS = zip(*[tuple(_.split(",")) for _ in sys.argv[1:]])  # indexed folder trees to search
  else:
    rootFolder = findRootFolder(None, constants.CONFIG)
    if rootFolder is None: error("Cannot automatically determine repository path. Please provide a list of REPOPATH,CONFIGPATH pairs as script arguments"); sys.exit(1)
    REPOS, METAS = [os.path.dirname(rootFolder)], [os.path.dirname(rootFolder)]

  idx = [Indexer(repo) for repo in REPOS]  # create index objects with location of storage
  for r, repo in enumerate(idx): repo.index = METAS[r]  # set location of metadata
  for r, repo in enumerate(idx): repo.load(os.path.join(METAS[r], constants.INDEX), True, False)  # load indexes
  for repo in idx: logging.info("Pre-filling path cache for %s" % repo.root); repo.findFolders([], [], True)  # trigger path cache filling - TODO hacky...
  hashcache = set()  # define global variable
  server = HTTPServer(SERVER, Handler)
  info("Starting HTTP server loop")
  while True: server.handle_request()
